@@ -2,6 +2,7 @@
  * PicoDrive
  * (c) Copyright Dave, 2004
  * (C) notaz, 2006-2010
+ * (C) irixxxx, 2020-2024
  *
  * This work is licensed under the terms of MAME license.
  * See COPYING file in the top-level directory.
@@ -796,12 +797,6 @@ int PicoCartLoad(pm_file *f, const unsigned char *rom, unsigned int romsize,
 
   if (!is_sms)
   {
-    // maybe we are loading MegaCD BIOS?
-    if (!(PicoIn.AHW & PAHW_MCD) && size == 0x20000 && (!strncmp((char *)rom_data+0x124, "BOOT", 4) ||
-         !strncmp((char *)rom_data+0x128, "BOOT", 4))) {
-      PicoIn.AHW |= PAHW_MCD;
-    }
-
     // Check for SMD:
     if (size >= 0x4200 && (size&0x3fff) == 0x200 &&
         ((rom_data[0x2280] == 'S' && rom_data[0x280] == 'E') || (rom_data[0x280] == 'S' && rom_data[0x2281] == 'E'))) {
@@ -833,7 +828,7 @@ int PicoCartInsert(unsigned char *rom, unsigned int romsize, const char *carthw_
   // This will hang the emu, but will prevent nasty crashes.
   // note: 4 bytes are padded to every ROM
   if (rom != NULL)
-    *(u32 *)(rom+romsize) = CPU_BE2(0x4EFAFFFE);
+    *(u32 *)(rom+romsize) = CPU_BE2(0x6000FFFE);
 
   Pico.rom=rom;
   Pico.romsize=romsize;
@@ -858,7 +853,7 @@ int PicoCartInsert(unsigned char *rom, unsigned int romsize, const char *carthw_
   PicoLoadStateHook = NULL;
   carthw_chunks = NULL;
 
-  if (!(PicoIn.AHW & (PAHW_MCD|PAHW_SMS|PAHW_PICO)))
+  if (!(PicoIn.AHW & (PAHW_SMS|PAHW_PICO)))
     PicoCartDetect(carthw_cfg);
   if (PicoIn.AHW & PAHW_SMS)
     PicoCartDetectMS();
@@ -921,7 +916,7 @@ void PicoCartUnload(void)
 static unsigned int rom_crc32(int size)
 {
   unsigned int crc;
-  elprintf(EL_STATUS, "caclulating CRC32..");
+  elprintf(EL_STATUS, "calculating CRC32..");
   if (size <= 0 || size > Pico.romsize) size = Pico.romsize;
 
   // have to unbyteswap for calculation..
@@ -931,14 +926,18 @@ static unsigned int rom_crc32(int size)
   return crc;
 }
 
-static int rom_strcmp(int rom_offset, const char *s1)
+int rom_strcmp(void *rom, int size, int offset, const char *s1)
 {
   int i, len = strlen(s1);
-  const char *s_rom = (const char *)Pico.rom;
-  if (rom_offset + len > Pico.romsize)
-    return 0;
+  const char *s_rom = (const char *)rom;
+  if (offset + len > size)
+    return 1;
+
+  if (PicoIn.AHW & PAHW_SMS)
+    return strncmp(s_rom + offset, s1, strlen(s1));
+
   for (i = 0; i < len; i++)
-    if (s1[i] != s_rom[MEM_BE2(i + rom_offset)])
+    if (s1[i] != s_rom[MEM_BE2(i + offset)])
       return 1;
   return 0;
 }
@@ -1062,7 +1061,7 @@ static void parse_carthw(const char *carthw_cfg, int *fill_sram,
     {
       int offs;
       offs = strtoul(p, &r, 0);
-      if (offs < 0 || offs > Pico.romsize) {
+      if (offs < 0) {
         elprintf(EL_STATUS, "carthw:%d: check_str offs out of range: %d\n", line, offs);
 	goto bad;
       }
@@ -1078,7 +1077,7 @@ static void parse_carthw(const char *carthw_cfg, int *fill_sram,
         goto bad;
       *r = 0;
 
-      if (rom_strcmp(offs, p) == 0)
+      if (rom_strcmp(Pico.rom, Pico.romsize, offs, p) == 0)
         any_checks_passed = 1;
       else
         skip_sect = 1;
@@ -1137,8 +1136,12 @@ static void parse_carthw(const char *carthw_cfg, int *fill_sram,
         PicoIn.AHW = PAHW_SVP;
       else if (strcmp(p, "pico") == 0)
         PicoIn.AHW = PAHW_PICO;
+      else if (strcmp(p, "j_cart") == 0)
+        carthw_jcart_startup();
       else if (strcmp(p, "prot") == 0)
         carthw_sprot_startup();
+      else if (strcmp(p, "flash") == 0)
+        carthw_flash_startup();
       else if (strcmp(p, "ssf2_mapper") == 0)
         carthw_ssf2_startup();
       else if (strcmp(p, "x_in_1_mapper") == 0)
@@ -1341,10 +1344,6 @@ static void PicoCartDetect(const char *carthw_cfg)
     memset(Pico.sv.data, 0xff, Pico.sv.size);
   }
 
-  // Unusual region 'code'
-  if (rom_strcmp(0x1f0, "EUROPE") == 0 || rom_strcmp(0x1f0, "Europe") == 0)
-    *(u32 *) (Pico.rom + 0x1f0) = CPU_LE4(0x20204520);
-
   // tweak for Blackthorne: master SH2 overwrites stack of slave SH2 being in PWM
   // interrupt. On real hardware, nothing happens since slave fetches the values
   // it has written from its cache, but picodrive doesn't emulate caching.
