@@ -153,8 +153,13 @@ static void ogg_free(int i)
 #endif
 #endif
 
+static off_t read_pos = -1;
+
 void cdd_reset(void)
 {
+  /* stop audio streaming */
+  Pico_mcd->cdda_stream = NULL;
+
   /* reset cycle counter */
   cdd.cycles = 0;
   
@@ -175,18 +180,29 @@ void cdd_reset(void)
 
   /* clear CD-DA output */
   cdd.audio[0] = cdd.audio[1] = 0;
+
+  /* no audio track playing */
+  Pico_mcd->s68k_regs[0x36+0] = 0x01;
+  /* reset file read position */
+  read_pos = -1;
 }
 
 /* FIXME: use cdd_read_audio() instead */
-static void cdd_change_track(int index, int lba)
+void cdd_play_audio(int index, int lba)
 {
   int i, base, lba_offset, lb_len;
 
   for (i = index; i >= 0; i--)
     if (cdd.toc.tracks[i].fd != NULL)
       break;
+  // prevent playing from file with binary track if MD+ (Doom Fusion)
+  // TODO: this doesn't cover all tracks being in a single bin file properly:
+  // in that case, fd should be duplicated to work properly with this MD+ shit.
+  if (! is_audio(index) || (read_pos >= 0 && cdd.index == i && ! is_audio(i)))
+    return;
 
   Pico_mcd->cdda_stream = cdd.toc.tracks[i].fd;
+  Pico_mcd->cdda_type = cdd.toc.tracks[i].type;
   base = cdd.toc.tracks[index].offset;
   lba_offset = lba - cdd.toc.tracks[index].start;
   lb_len = cdd.toc.tracks[index].end - cdd.toc.tracks[index].start;
@@ -196,6 +212,63 @@ static void cdd_change_track(int index, int lba)
   cdda_start_play(base, lba_offset, lb_len);
 }
 
+static void cdd_seek(int index, int lba)
+{
+#ifdef USE_LIBTREMOR
+#ifdef DISABLE_MANY_OGG_OPEN_FILES
+  /* close previous track VORBIS file structure to save memory */
+  if (is_audio(cdd.index) && cdd.toc.tracks[cdd.index].vf.datasource)
+  {
+    ogg_free(cdd.index);
+  }
+
+  /* open current track VORBIS file */
+  if (is_audio(index) && cdd.toc.tracks[index].vf.seekable)
+  {
+    ov_open(cdd.toc.tracks[index].fd,&cdd.toc.tracks[index].vf,0,0);
+  }
+#endif
+#endif
+
+  /* update current track index and LBA */
+  cdd.index = (index < 0 ? 0 : index);
+  cdd.lba = lba;
+  if (index < 0) return;
+
+  /* stay within track limits when seeking files */
+  if (lba < cdd.toc.tracks[cdd.index].start)
+  {
+    lba = cdd.toc.tracks[cdd.index].start;
+  }
+
+  /* seek to current block */
+  if (!is_audio(cdd.index))
+  {
+    /* DATA track */
+    read_pos = lba * cdd.sectorSize;
+    pm_seek(cdd.toc.tracks[cdd.index].fd, read_pos, SEEK_SET);
+  }
+#ifdef USE_LIBTREMOR
+  else if (cdd.toc.tracks[cdd.index].vf.seekable)
+  {
+    /* VORBIS AUDIO track */
+    ov_pcm_seek(&cdd.toc.tracks[cdd.index].vf, (lba - cdd.toc.tracks[cdd.index].start) * 588 - cdd.toc.tracks[cdd.index].offset);
+  }
+#endif
+#if 0
+  else if (cdd.toc.tracks[cdd.index].fd)
+  {
+    /* PCM AUDIO track */
+    fseek(cdd.toc.tracks[cdd.index].fd, (lba * 2352) - cdd.toc.tracks[cdd.index].offset, SEEK_SET);
+  }
+#else
+  else
+  {
+    cdd_play_audio(cdd.index, lba);
+  }
+#endif
+}
+
 int cdd_context_save(uint8 *state)
 {
   int bufferptr = 0;
@@ -213,19 +286,8 @@ int cdd_context_save(uint8 *state)
 
 int cdd_context_load(uint8 *state)
 {
-  int lba;
   int bufferptr = 0;
 
-#ifdef USE_LIBTREMOR
-#ifdef DISABLE_MANY_OGG_OPEN_FILES
-  /* close previous track VORBIS file structure to save memory */
-  if (cdd.toc.tracks[cdd.index].vf.datasource)
-  {
-    ogg_free(cdd.index);
-  }
-#endif
-#endif
-
   load_param(&cdd.cycles, sizeof(cdd.cycles));
   load_param(&cdd.latency, sizeof(cdd.latency));
   load_param(&cdd.index, sizeof(cdd.index));
@@ -234,45 +296,8 @@ int cdd_context_load(uint8 *state)
   load_param(&cdd.volume, sizeof(cdd.volume));
   load_param(&cdd.status, sizeof(cdd.status));
 
-  /* adjust current LBA within track limit */
-  lba = cdd.lba;
-  if (lba < cdd.toc.tracks[cdd.index].start)
-  {
-    lba = cdd.toc.tracks[cdd.index].start;
-  }
-
   /* seek to current track position */
-  if (!is_audio(cdd.index))
-  {
-    /* DATA track */
-    if (cdd.toc.tracks[cdd.index].fd)
-    {
-      pm_seek(cdd.toc.tracks[cdd.index].fd, lba * cdd.sectorSize, SEEK_SET);
-    }
-  }
-#ifdef USE_LIBTREMOR
-  else if (cdd.toc.tracks[cdd.index].vf.seekable)
-  {
-#ifdef DISABLE_MANY_OGG_OPEN_FILES
-    /* VORBIS file need to be opened first */
-    ov_open(cdd.toc.tracks[cdd.index].fd,&cdd.toc.tracks[cdd.index].vf,0,0);
-#endif
-    /* VORBIS AUDIO track */
-    ov_pcm_seek(&cdd.toc.tracks[cdd.index].vf, (lba - cdd.toc.tracks[cdd.index].start) * 588 - cdd.toc.tracks[cdd.index].offset);
-  }
-#endif
-#if 0
-  else if (cdd.toc.tracks[cdd.index].fd)
-  {
-    /* PCM AUDIO track */
-    fseek(cdd.toc.tracks[cdd.index].fd, (lba * 2352) - cdd.toc.tracks[cdd.index].offset, SEEK_SET);
-  }
-#else
-  else
-  {
-    cdd_change_track(cdd.index, lba);
-  }
-#endif
+  cdd_seek(cdd.index, cdd.lba);
 
   return bufferptr;
 }
@@ -280,6 +305,8 @@ int cdd_context_load(uint8 *state)
 int cdd_context_load_old(uint8 *state)
 {
   memcpy(&cdd.lba, state + 8, sizeof(cdd.lba));
+  cdd_seek(cdd.index, cdd.lba);
+
   return 12 * 4;
 }
 
@@ -296,12 +323,20 @@ int cdd_load(const char *filename, int type)
   if (ret != 0)
     return ret;
 
+  if (type == CT_ISO) {
+    /* ISO format (2048 bytes data blocks) */
+    cdd.sectorSize = 2048;
+  } else {
+    /* audio or BIN format (2352 bytes data blocks) */
+    cdd.sectorSize = 2352;
+  }
+
   /* read first 16 bytes */
   pm_read(header, 0x10, cdd.toc.tracks[0].fd);
 
   /* look for valid CD image ID string */
-  if (memcmp("SEGADISCSYSTEM", header, 14))
-  {    
+  if (!Pico.romsize && memcmp("SEGADISCSYSTEM", header, 14))
+  {
     /* if not found, read next 16 bytes */
     pm_read(header, 0x10, cdd.toc.tracks[0].fd);
 
@@ -311,27 +346,16 @@ int cdd_load(const char *filename, int type)
       elprintf(EL_STATUS|EL_ANOMALY, "cd: bad cd image?");
       /* assume bin without security code */
     }
-
-    /* BIN format (2352 bytes data blocks) */
-    cdd.sectorSize = 2352;
-  }
-  else
-  {
-    /* ISO format (2048 bytes data blocks) */
-    cdd.sectorSize = 2048;
   }
 
-  ret = (type == CT_ISO ? 2048 : 2352);
-  if (ret != cdd.sectorSize)
-    elprintf(EL_STATUS|EL_ANOMALY, "cd: type detection mismatch");
   pm_sectorsize(cdd.sectorSize, cdd.toc.tracks[0].fd);
 
-  /* read CD image header + security code */
-  pm_read(header + 0x10, 0x200, cdd.toc.tracks[0].fd);
-
   /* Simulate audio tracks if none found */
-  if (cdd.toc.last == 1)
+  if (!Pico.romsize && cdd.toc.last == 1)
   {
+    /* read CD image header + security code */
+    pm_read(header + 0x10, 0x200, cdd.toc.tracks[0].fd);
+
     /* Some games require exact TOC infos */
     if (strstr(header + 0x180,"T-95035") != NULL)
     {
@@ -448,7 +472,7 @@ int cdd_unload(void)
     int i;
 
     /* stop audio streaming */
-    Pico_mcd->cdda_stream = NULL;
+    if (Pico_mcd) Pico_mcd->cdda_stream = NULL;
 
     /* close CD tracks */
     if (cdd.toc.tracks[0].fd)
@@ -473,7 +497,7 @@ int cdd_unload(void)
       if (cdd.toc.tracks[i].fd)
       {
         /* close file */
-        if (Pico_mcd->cdda_type == CT_MP3)
+        if (cdd.toc.tracks[i].type == CT_MP3)
           fclose(cdd.toc.tracks[i].fd);
         else
           pm_close(cdd.toc.tracks[i].fd);
@@ -510,17 +534,29 @@ int cdd_unload(void)
 void cdd_read_data(uint8 *dst)
 {
   /* only read DATA track sectors */
-  if (!is_audio(cdd.index) && (cdd.lba >= 0) && (cdd.lba < cdd.toc.tracks[cdd.index].end))
+  if (!is_audio(cdd.index) && (cdd.lba >= cdd.toc.tracks[cdd.index].start) &&
+                              (cdd.lba < cdd.toc.tracks[cdd.index].end))
   {
+    off_t pos;
+
     /* BIN format ? */
     if (cdd.sectorSize == 2352)
     {
       /* skip 16-byte header */
-      pm_seek(cdd.toc.tracks[cdd.index].fd, cdd.lba * 2352 + 16, SEEK_SET);
+      pos = cdd.lba * 2352 + 16;
+    }
+    else
+    {
+      pos = cdd.lba * cdd.sectorSize;
+    }
+
+    if (pos != read_pos) {
+      pm_seek(cdd.toc.tracks[cdd.index].fd, pos, SEEK_SET);
+      read_pos = pos;
     }
 
     /* read sector data (Mode 1 = 2048 bytes) */
-    pm_read(dst, 2048, cdd.toc.tracks[cdd.index].fd);
+    read_pos += pm_read(dst, 2048, cdd.toc.tracks[cdd.index].fd);
   }
 }
 
@@ -685,11 +721,39 @@ void cdd_read_audio(unsigned int samples)
 
 
 void cdd_update(void)
-{  
+{
 #ifdef LOG_CDD
   error("LBA = %d (track n%d)(latency=%d)\n", cdd.lba, cdd.index, cdd.latency);
 #endif
   
+  /* update decoder, depending on track type */
+  if (cdd.status == CD_PLAY && !is_audio(cdd.index) && !cdd.latency)
+  {
+    /* DATA sector header (CD-ROM Mode 1) */
+    uint8 header[4];
+    uint32 msf = cdd.lba + 150;
+    header[0] = lut_BCD_8[(msf / 75) / 60];
+    header[1] = lut_BCD_8[(msf / 75) % 60];
+    header[2] = lut_BCD_8[(msf % 75)];
+    header[3] = 0x01;
+
+    /* data track sector read is controlled by CDC */
+    cdc_decoder_update(header);
+  }
+  else
+  {
+    uint8 header[4] = { 0, };
+
+    /* audio blocks are still sent to CDC as well as CD DAC/Fader */
+    cdc_decoder_update(header);
+  }
+
+  if (Pico_mcd->m.state_flags & PCD_ST_CDD_CMD) {
+    /* pending delayed command */
+    cdd_process();
+    Pico_mcd->m.state_flags &= ~PCD_ST_CDD_CMD;
+  }
+
   /* drive latency */
   if (cdd.latency > 0)
   {
@@ -707,78 +771,24 @@ void cdd_update(void)
       return;
     }
 
-    /* track type */
-    if (!is_audio(cdd.index))
+    /* check against audio track start index */
+    if (is_audio(cdd.index) && cdd.lba >= cdd.toc.tracks[cdd.index].start)
     {
-      /* DATA sector header (CD-ROM Mode 1) */
-      uint8 header[4];
-      uint32 msf = cdd.lba + 150;
-      header[0] = lut_BCD_8[(msf / 75) / 60];
-      header[1] = lut_BCD_8[(msf / 75) % 60];
-      header[2] = lut_BCD_8[(msf % 75)];
-      header[3] = 0x01;
-
-      /* data track sector read is controlled by CDC */
-      cdd.lba += cdc_decoder_update(header);
+      /* audio track playing */
+      Pico_mcd->s68k_regs[0x36+0] = 0x00;
     }
-    else
-    {
-      uint8 header[4] = { 0, };
-
-      /* check against audio track start index */
-      if (cdd.lba >= cdd.toc.tracks[cdd.index].start)
-      {
-        /* audio track playing */
-        Pico_mcd->s68k_regs[0x36+0] = 0x00;
-      }
-
-      /* audio blocks are still sent to CDC as well as CD DAC/Fader */
-      cdc_decoder_update(header);
  
-      /* next audio block is automatically read */
-      cdd.lba++;
-    }
+    /* next block is automatically read */
+    cdd.lba++;
 
     /* check end of current track */
     if (cdd.lba >= cdd.toc.tracks[cdd.index].end)
     {
-#ifdef USE_LIBTREMOR
-#ifdef DISABLE_MANY_OGG_OPEN_FILES
-      /* close previous track VORBIS file structure to save memory */
-      if (cdd.toc.tracks[cdd.index].vf.datasource)
-      {
-        ogg_free(cdd.index);
-      }
-#endif
-#endif
-      /* play next track */
-      cdd.index++;
-
       /* PAUSE between tracks */
       Pico_mcd->s68k_regs[0x36+0] = 0x01;
 
       /* seek to next audio track start */
-#ifdef USE_LIBTREMOR
-      if (cdd.toc.tracks[cdd.index].vf.seekable)
-      {
-#ifdef DISABLE_MANY_OGG_OPEN_FILES
-        /* VORBIS file need to be opened first */
-        ov_open(cdd.toc.tracks[cdd.index].fd,&cdd.toc.tracks[cdd.index].vf,0,0);
-#endif
-        ov_pcm_seek(&cdd.toc.tracks[cdd.index].vf, -cdd.toc.tracks[cdd.index].offset);
-      }
-      else
-#endif 
-#if 0
-      if (cdd.toc.tracks[cdd.index].fd)
-      {
-        fseek(cdd.toc.tracks[cdd.index].fd, (cdd.toc.tracks[cdd.index].start * 2352) - cdd.toc.tracks[cdd.index].offset, SEEK_SET);
-      }
-#else
-      {
-        cdd_change_track(cdd.index, cdd.lba);
-      }
-#endif
+      cdd_seek(cdd.index + 1, cdd.lba);
     }
   }
 
@@ -791,100 +801,45 @@ void cdd_update(void)
     /* check current track limits */
     if (cdd.lba >= cdd.toc.tracks[cdd.index].end)
     {
-#ifdef USE_LIBTREMOR
-#ifdef DISABLE_MANY_OGG_OPEN_FILES
-      /* close previous track VORBIS file structure to save memory */
-      if (cdd.toc.tracks[cdd.index].vf.datasource)
-      {
-        ogg_free(cdd.index);
-      }
-#endif
-#endif
       /* next track */
-      cdd.index++;
-
-      /* skip directly to track start position */
-      cdd.lba = cdd.toc.tracks[cdd.index].start;
-      
-      /* AUDIO track playing ? */
-      if (cdd.status == CD_PLAY)
+      if (cdd.index >= cdd.toc.last)
       {
-        Pico_mcd->s68k_regs[0x36+0] = 0x00;
+        /* no AUDIO track playing */
+        Pico_mcd->s68k_regs[0x36+0] = 0x01;
+
+        /* end of disc */
+        cdd.lba = cdd.toc.end;
+        cdd.status = CD_END;
+      }
+      else
+      {
+        cdd_seek(cdd.index + 1, cdd.toc.tracks[cdd.index+1].start);
+
+        /* AUDIO track playing ? */
+        if (cdd.status == CD_PLAY && is_audio(cdd.index))
+        {
+          Pico_mcd->s68k_regs[0x36+0] = 0x00;
+        }
       }
     }
     else if (cdd.lba < cdd.toc.tracks[cdd.index].start)
     {
-#ifdef USE_LIBTREMOR
-#ifdef DISABLE_MANY_OGG_OPEN_FILES
-      /* close previous track VORBIS file structure to save memory */
-      if (cdd.toc.tracks[cdd.index].vf.datasource)
-      {
-        ogg_free(cdd.index);
-      }
-#endif
-#endif
-
       /* previous track */
-      cdd.index--;
-
-      /* skip directly to track end position */
-      cdd.lba = cdd.toc.tracks[cdd.index].end;
+      if (cdd.index <= 0)
+      {
+        cdd_seek(0, 0);
+      }
+      else
+      {
+        cdd_seek(cdd.index - 1, cdd.toc.tracks[cdd.index-1].end);
+      }
     }
 
-    /* check disc limits */
-    if (cdd.index < 0)
-    {
-      cdd.index = 0;
-      cdd.lba = 0;
-    }
-    else if (cdd.index >= cdd.toc.last)
-    {
-      /* no AUDIO track playing */
-      Pico_mcd->s68k_regs[0x36+0] = 0x01;
-
-      /* end of disc */
-      cdd.index = cdd.toc.last;
-      cdd.lba = cdd.toc.end;
-      cdd.status = CD_END;
-      return;
-    }
-
-    /* seek to current block */
     if (!is_audio(cdd.index))
     {
       /* no AUDIO track playing */
       Pico_mcd->s68k_regs[0x36+0] = 0x01;
-
-      /* DATA track */
-      pm_seek(cdd.toc.tracks[cdd.index].fd, cdd.lba * cdd.sectorSize, SEEK_SET);
     }
-#ifdef USE_LIBTREMOR
-    else if (cdd.toc.tracks[cdd.index].vf.seekable)
-    {
-#ifdef DISABLE_MANY_OGG_OPEN_FILES
-      /* check if a new track is being played */
-      if (!cdd.toc.tracks[cdd.index].vf.datasource)
-      {
-        /* VORBIS file need to be opened first */
-        ov_open(cdd.toc.tracks[cdd.index].fd,&cdd.toc.tracks[cdd.index].vf,0,0);
-      }
-#endif
-      /* VORBIS AUDIO track */
-      ov_pcm_seek(&cdd.toc.tracks[cdd.index].vf, (cdd.lba - cdd.toc.tracks[cdd.index].start) * 588 - cdd.toc.tracks[cdd.index].offset);
-    }
-#endif 
-#if 0
-    else if (cdd.toc.tracks[cdd.index].fd)
-    {
-      /* PCM AUDIO track */
-      fseek(cdd.toc.tracks[cdd.index].fd, (cdd.lba * 2352) - cdd.toc.tracks[cdd.index].offset, SEEK_SET);
-    }
-#else
-    else
-    {
-      cdd_change_track(cdd.index, cdd.lba);
-    }
-#endif
   }
 }
 
@@ -901,7 +856,7 @@ void cdd_process(void)
   {
     case 0x00:  /* Drive Status */
     {
-      if (cdd.latency <= 3) {
+      if (cdd.latency == 0) {
         /* RS1-RS8 normally unchanged */
         Pico_mcd->s68k_regs[0x38+0] = cdd.status;
 
@@ -910,7 +865,7 @@ void cdd_process(void)
             Pico_mcd->s68k_regs[0x38+1] == 0x00 ||
             Pico_mcd->s68k_regs[0x38+1] == 0x01)
         {
-          int lba = cdd.lba + 150;
+          int lba = cdd.lba + 150 - cdd.latency;
 	  if (Pico_mcd->s68k_regs[0x38+1] == 0x01)
             lba = abs(cdd.lba - cdd.toc.tracks[cdd.index].start);
 	  if (Pico_mcd->s68k_regs[0x38+1] == 0x0f)
@@ -942,6 +897,9 @@ void cdd_process(void)
       set_reg16(0x3c, 0x0000);
       set_reg16(0x3e, 0x0000);
       set_reg16(0x40, 0x000f);
+
+      /* reset to 1st track */
+      cdd_seek(0, 0);
       return;
     }
 
@@ -1056,6 +1014,19 @@ void cdd_process(void)
                  (Pico_mcd->s68k_regs[0x46+0] * 10 + Pico_mcd->s68k_regs[0x46+1])) * 75 +
                  (Pico_mcd->s68k_regs[0x48+0] * 10 + Pico_mcd->s68k_regs[0x48+1]) - 150;
 
+      /* return track index in RS2-RS3 */
+      set_reg16(0x38, (CD_SEEK << 8) | 0x0f);
+      set_reg16(0x3a, 0x0000);
+      set_reg16(0x3c, 0x0000);
+      set_reg16(0x3e, 0x0000);
+      set_reg16(0x40, ~(CD_SEEK + 0xf) & 0x0f);
+
+      /* seek delayed by max 2 blocks before the seek starts */
+      if (!(Pico_mcd->m.state_flags & PCD_ST_CDD_CMD)) {
+        Pico_mcd->m.state_flags |= PCD_ST_CDD_CMD;
+        return;
+      }
+
       /* CD drive latency */
       if (!cdd.latency)
       {
@@ -1081,66 +1052,11 @@ void cdd_process(void)
         cdd.latency += (((cdd.lba - lba) * 120) / 270000);
       }
 
-      /* update current LBA */
-      cdd.lba = lba;
-
       /* get track index */
       while ((cdd.toc.tracks[index].end <= lba) && (index < cdd.toc.last)) index++;
 
-#ifdef USE_LIBTREMOR
-#ifdef DISABLE_MANY_OGG_OPEN_FILES
-      /* check if track index has changed */
-      if (index != cdd.index)
-      {
-        /* close previous track VORBIS file structure to save memory */
-        if (cdd.toc.tracks[cdd.index].vf.datasource)
-        {
-          ogg_free(cdd.index);
-        }
-
-        /* open current track VORBIS file */
-        if (cdd.toc.tracks[index].vf.seekable)
-        {
-          ov_open(cdd.toc.tracks[index].fd,&cdd.toc.tracks[index].vf,0,0);
-        }
-      }
-#endif
-#endif
-
-      /* update current track index */
-      cdd.index = index;
-
-      /* stay within track limits when seeking files */
-      if (lba < cdd.toc.tracks[index].start) 
-      {
-        lba = cdd.toc.tracks[index].start;
-      }
-      
-      /* seek to current block */
-      if (!is_audio(cdd.index))
-      {
-        /* DATA track */
-        pm_seek(cdd.toc.tracks[cdd.index].fd, lba * cdd.sectorSize, SEEK_SET);
-      }
-#ifdef USE_LIBTREMOR
-      else if (cdd.toc.tracks[index].vf.seekable)
-      {
-        /* VORBIS AUDIO track */
-        ov_pcm_seek(&cdd.toc.tracks[index].vf, (lba - cdd.toc.tracks[index].start) * 588 - cdd.toc.tracks[index].offset);
-      }
-#endif
-#if 0
-      else if (cdd.toc.tracks[index].fd)
-      {
-        /* PCM AUDIO track */
-        fseek(cdd.toc.tracks[index].fd, (lba * 2352) - cdd.toc.tracks[index].offset, SEEK_SET);
-      }
-#else
-      else
-      {
-        cdd_change_track(index, lba);
-      }
-#endif
+      /* seek to block; playing always starts 3 blocks earlier */
+      cdd_seek(index, lba - 3);
 
       /* no audio track playing (yet) */
       Pico_mcd->s68k_regs[0x36+0] = 0x01;
@@ -1148,13 +1064,7 @@ void cdd_process(void)
       /* update status */
       cdd.status = CD_PLAY;
 
-      /* return track index in RS2-RS3 */
-      set_reg16(0x38, (CD_SEEK << 8) | 0x0f);
-      set_reg16(0x3a, 0x0000);
-      set_reg16(0x3c, 0x0000);
-      set_reg16(0x3e, 0x0000);
-      set_reg16(0x40, ~(CD_SEEK + 0xf) & 0x0f);
-      break;
+      return;
     }
 
     case 0x04:  /* Seek */
@@ -1167,6 +1077,19 @@ void cdd_process(void)
                  (Pico_mcd->s68k_regs[0x46+0] * 10 + Pico_mcd->s68k_regs[0x46+1])) * 75 +
                  (Pico_mcd->s68k_regs[0x48+0] * 10 + Pico_mcd->s68k_regs[0x48+1]) - 150;
 
+      /* unknown RS1-RS8 values (returning 0xF in RS1 invalidates track infos in RS2-RS8 and fixes Final Fight CD intro when seek time is emulated) */
+      set_reg16(0x38, (CD_SEEK << 8) | 0x0f);
+      set_reg16(0x3a, 0x0000);
+      set_reg16(0x3c, 0x0000);
+      set_reg16(0x3e, 0x0000);
+      set_reg16(0x40, ~(CD_SEEK + 0xf) & 0x0f);
+
+      /* seek delayed by max 2 blocks before the seek starts */
+      if (!(Pico_mcd->m.state_flags & PCD_ST_CDD_CMD)) {
+        Pico_mcd->m.state_flags |= PCD_ST_CDD_CMD;
+        return;
+      }
+
       /* CD drive seek time  */
       /* We are using similar linear model as above, although still not exactly accurate, */
       /* it works fine for Switch/Panic! intro (Switch needs at least 30 interrupts while */
@@ -1180,66 +1103,11 @@ void cdd_process(void)
         cdd.latency = ((cdd.lba - lba) * 120) / 270000;
       }
 
-      /* update current LBA */
-      cdd.lba = lba;
-
-      /* get current track index */
+      /* get track index */
       while ((cdd.toc.tracks[index].end <= lba) && (index < cdd.toc.last)) index++;
 
-#ifdef USE_LIBTREMOR
-#ifdef DISABLE_MANY_OGG_OPEN_FILES
-      /* check if track index has changed */
-      if (index != cdd.index)
-      {
-        /* close previous track VORBIS file structure to save memory */
-        if (cdd.toc.tracks[cdd.index].vf.datasource)
-        {
-          ogg_free(cdd.index);
-        }
-
-        /* open current track VORBIS file */
-        if (cdd.toc.tracks[index].vf.seekable)
-        {
-          ov_open(cdd.toc.tracks[index].fd,&cdd.toc.tracks[index].vf,0,0);
-        }
-      }
-#endif
-#endif
-
-      /* update current track index */
-      cdd.index = index;
-
-      /* stay within track limits */
-      if (lba < cdd.toc.tracks[index].start) 
-      {
-        lba = cdd.toc.tracks[index].start;
-      }
-      
-      /* seek to current block */
-      if (!is_audio(cdd.index))
-      {
-        /* DATA track */
-        pm_seek(cdd.toc.tracks[cdd.index].fd, lba * cdd.sectorSize, SEEK_SET);
-      }
-#ifdef USE_LIBTREMOR
-      else if (cdd.toc.tracks[index].vf.seekable)
-      {
-        /* VORBIS AUDIO track */
-        ov_pcm_seek(&cdd.toc.tracks[index].vf, (lba - cdd.toc.tracks[index].start) * 588 - cdd.toc.tracks[index].offset);
-      }
-#endif
-#if 0
-      else if (cdd.toc.tracks[index].fd)
-      {
-        /* PCM AUDIO track */
-        fseek(cdd.toc.tracks[index].fd, (lba * 2352) - cdd.toc.tracks[index].offset, SEEK_SET);
-      }
-#else
-      else
-      {
-        cdd_change_track(index, lba);
-      }
-#endif
+      /* seek to block; playing always starts 3 blocks earlier */
+      cdd_seek(index, lba - 3);
 
       /* no audio track playing */
       Pico_mcd->s68k_regs[0x36+0] = 0x01;
@@ -1247,22 +1115,17 @@ void cdd_process(void)
       /* update status */
       cdd.status = CD_READY;
 
-      /* unknown RS1-RS8 values (returning 0xF in RS1 invalidates track infos in RS2-RS8 and fixes Final Fight CD intro when seek time is emulated) */
-      set_reg16(0x38, (CD_SEEK << 8) | 0x0f);
-      set_reg16(0x3a, 0x0000);
-      set_reg16(0x3c, 0x0000);
-      set_reg16(0x3e, 0x0000);
-      set_reg16(0x40, ~(CD_SEEK + 0xf) & 0x0f);
       return;
     }
 
     case 0x06:  /* Pause */
     {
+      /* update status (RS1-RS8 unchanged) */
+      cdd.status = Pico_mcd->s68k_regs[0x38+0] = CD_READY;
+
       /* no audio track playing */
       Pico_mcd->s68k_regs[0x36+0] = 0x01;
 
-      /* update status (RS1-RS8 unchanged) */
-      cdd.status = Pico_mcd->s68k_regs[0x38+0] = CD_READY;
       break;
     }
 
