@@ -2,13 +2,16 @@
  * memory handling
  * (c) Copyright Dave, 2004
  * (C) notaz, 2006-2010
+ * (C) irixxxx, 2019-2024
  *
  * This work is licensed under the terms of MAME license.
  * See COPYING file in the top-level directory.
  */
 
+#include <assert.h>
 #include "pico_int.h"
 #include "memory.h"
+#include "state.h"
 
 #include "sound/ym2612.h"
 #include "sound/sn76496.h"
@@ -24,9 +27,9 @@ static void xmap_set(uptr *map, int shift, u32 start_addr, u32 end_addr,
     const void *func_or_mh, int is_func)
 {
 #ifdef __clang__
-  // workaround bug (segfault) in 
+  // workaround bug (segfault) in
   // Apple LLVM version 4.2 (clang-425.0.27) (based on LLVM 3.2svn)
-  volatile 
+  volatile
 #endif
   uptr addr = (uptr)func_or_mh;
   int mask = (1 << shift) - 1;
@@ -202,22 +205,22 @@ void cpu68k_map_all_funcs(u32 start_addr, u32 end_addr, u32 (*r8)(u32), u32 (*r1
 u32 PicoRead16_floating(u32 a)
 {
   // faking open bus
-  u32 d = (Pico.m.rotate += 0x41);
+  u16 d = (Pico.m.rotate += 0x41);
   d ^= (d << 5) ^ (d << 8);
-  if ((a & 0xff0000) == 0xa10000) d = 0; // MegaCD pulldowns don't work here curiously
+  if ((a & 0xff0000) == 0xa10000) return d; // MegaCD pulldowns don't work here curiously
   return (PicoIn.AHW & PAHW_MCD) ? 0x00 : d; // pulldown if MegaCD2 attached
 }
 
 static u32 m68k_unmapped_read8(u32 a)
 {
   elprintf(EL_UIO, "m68k unmapped r8  [%06x] @%06x", a, SekPc);
-  return (u8)PicoRead16_floating(a);
+  return a < 0x400000 ? 0 : (u8)PicoRead16_floating(a);
 }
 
 static u32 m68k_unmapped_read16(u32 a)
 {
   elprintf(EL_UIO, "m68k unmapped r16 [%06x] @%06x", a, SekPc);
-  return PicoRead16_floating(a);
+  return a < 0x400000 ? 0 : PicoRead16_floating(a);
 }
 
 static void m68k_unmapped_write8(u32 a, u32 d)
@@ -233,9 +236,9 @@ static void m68k_unmapped_write16(u32 a, u32 d)
 void m68k_map_unmap(u32 start_addr, u32 end_addr)
 {
 #ifdef __clang__
-  // workaround bug (segfault) in 
+  // workaround bug (segfault) in
   // Apple LLVM version 4.2 (clang-425.0.27) (based on LLVM 3.2svn)
-  volatile 
+  volatile
 #endif
   uptr addr;
   int shift = M68K_MEM_SHIFT;
@@ -293,12 +296,22 @@ void log_io(unsigned int addr, int bits, int rw);
 #endif
 
 #if defined(EMU_C68K)
-void cyclone_crashed(u32 pc, struct Cyclone *context)
+u32 cyclone_crashed(u32 pc, struct Cyclone *context)
 {
-    elprintf(EL_STATUS|EL_ANOMALY, "%c68k crash detected @ %06x",
-      context == &PicoCpuCM68k ? 'm' : 's', pc);
+    // check for underlying ROM, in case of on-cart hw overlaying part of ROM
+    // NB assumes code isn't executed from the overlay, but I've never seen this
+    u32 pc24 = pc & 0xffffff;
+    if (pc24 >= Pico.romsize) {
+      // no ROM, so it's probably an illegal access
+      pc24 = Pico.romsize;
+      elprintf(EL_STATUS|EL_ANOMALY, "%c68k crash detected @ %06x",
+        context == &PicoCpuCM68k ? 'm' : 's', pc);
+    }
+
     context->membase = (u32)Pico.rom;
-    context->pc = (u32)Pico.rom + Pico.romsize;
+    context->pc = (u32)Pico.rom + pc24;
+
+    return context->pc;
 }
 #endif
 
@@ -329,7 +342,7 @@ static u32 read_pad_6btn(int i, u32 out_bits)
     value = (pad & 0xc0) >> 2;                   // ?0SA 0000
     goto out;
   }
-  else if(phase == 3) {
+  else if (phase == 3) {
     if (out_bits & 0x40)
       value = (pad & 0x30) | ((pad >> 8) & 0xf); // ?1CB MXYZ
     else
@@ -392,6 +405,82 @@ static u32 read_pad_4way(int i, u32 out_bits)
   return value;
 }
 
+static u32 read_pad_mouse(int i, u32 out_bits)
+{
+  int phase = Pico.m.padTHPhase[i];
+  u32 value;
+
+  int x = PicoIn.mouseInt[0] - PicoIn.mouseInt[2];
+  int y = PicoIn.mouseInt[3] - PicoIn.mouseInt[1];
+
+  switch (phase) {
+  case 0:
+    value = 0x00;
+    break;
+  case 1:
+    value = 0x0b;
+    break;
+  case 2:
+    value = 0x0f;
+    // store last read values for x,y difference calculation
+    PicoIn.mouseInt[2] = PicoIn.mouseInt[0];
+    PicoIn.mouseInt[3] = PicoIn.mouseInt[1];
+    break;
+  case 3:
+    value = 0x0f;
+    // latch current mouse position during readout
+    PicoIn.mouseInt[0] = PicoIn.mouse[0];
+    PicoIn.mouseInt[1] = PicoIn.mouse[1];
+    break;
+  case 4: // xxxx OOSS, OO = y,x overflow, SS = y,x sign bits
+    value = (x<0) | ((y<0)<<1) | ((x<-255||x>255)<<2) | ((y<-255||y>255)<<3);
+    break;
+  case 5:
+    value = (PicoIn.padInt[i] & 0xf0) >> 4;      // SMRL, mapped from SACB
+    break;
+  case 6: // high nibble of x
+    value = (x>>4) & 0xf;
+    break;
+  case 7: // low nibble of x
+    value = x & 0xf;
+    break;
+  case 8: // high nibble of y
+    value = (y>>4) & 0xf;
+    break;
+  case 9: // low nibble of y
+  default: // also sent on all later phases
+    value = y & 0xf;
+    break;
+  }
+
+  value |= (out_bits & 0x40) | ((out_bits & 0x20)>>1);
+  return value;
+}
+
+static u32 read_pad_menacer(int i, u32 out_bits)
+{
+  u32 pad = PicoIn.padInt[i]; // Get pad MXYZ SACB RLDU; NB not(!) inverted
+  u32 value;
+
+  value = ((pad>>4) & 0x09) | ((pad>>3) & 0x4) | ((pad>>5) & 0x2); // .... SCAB
+
+  value |= 0x40; // TH line is light sensor. Pretend there's always darkness...
+  return value;
+}
+
+static u32 read_pad_justifier(int i, u32 out_bits)
+{
+  u32 pad = ~PicoIn.padInt[i]; // Get inverse of pad MXYZ SACB RLDU
+  u32 value = 0x30; // .111 0000, for ID
+
+  // output data: .tgd ...., t = TH, g = gun select, d = disable
+  if (!(PicoMem.ioports[i+1] & out_bits & 0x70)) // blue gun, enabled
+    value |= ((pad>>6) & 0x03); // .011 ..SA
+
+  value |= (out_bits & 0x40);
+  return value;
+}
+
 static u32 read_nothing(int i, u32 out_bits)
 {
   return 0xff;
@@ -399,73 +488,129 @@ static u32 read_nothing(int i, u32 out_bits)
 
 typedef u32 (port_read_func)(int index, u32 out_bits);
 
+int port_type[3] = {
+  PICO_INPUT_PAD_3BTN,
+  PICO_INPUT_PAD_3BTN,
+  PICO_INPUT_NOTHING
+};
+int port_lightgun;
+
 static port_read_func *port_readers[3] = {
   read_pad_3btn,
   read_pad_3btn,
   read_nothing
 };
 
-static int padTHLatency[3]; // TODO this should be in the save file structures
+static int padTHLatency[3];
+static int padTLLatency[3];
+static int padTHTimeout[3];
 
 static NOINLINE u32 port_read(int i)
 {
   u32 data_reg = PicoMem.ioports[i + 1];
   u32 ctrl_reg = PicoMem.ioports[i + 4] | 0x80;
+  u32 cycles = SekCyclesDone();
   u32 in, out;
 
   out = data_reg & ctrl_reg;
 
+  // controllers have a timeout on TH, they reset if not toggled for some time
+  if (CYCLES_GE(cycles, padTHTimeout[i]) && i < 2) {
+    padTHTimeout[i] = cycles + 0x70000000;
+    Pico.m.padTHPhase[i] = 0;
+  }
+
   // pull-ups: should be 0x7f, but Decap Attack has a bug where it temp.
   // disables output before doing TH-low read, so emulate RC filter for TH.
-  // Decap Attack reportedly doesn't work on Nomad but works on must
+  // Decap Attack reportedly doesn't work on Nomad but works on most
   // other MD revisions (different pull-up strength?).
   u32 mask = 0x3f;
-  if (CYCLES_GE(padTHLatency[i], SekCyclesDone()+100))
-    padTHLatency[i] = SekCyclesDone(); // kludge to cope with cycle wrap
-  if (CYCLES_GE(SekCyclesDone(), padTHLatency[i])) {
+  if (CYCLES_GE(cycles, padTHLatency[i])) {
     mask |= 0x40;
-    padTHLatency[i] = SekCyclesDone();
+    padTHLatency[i] = cycles;
   }
   out |= mask & ~ctrl_reg;
 
   in = port_readers[i](i, out);
 
+  // Sega mouse uses the TL/TR lines for req/ack. For buggy drivers, make sure
+  // there's some delay before ack is sent by taking over the new TL line level
+  if (CYCLES_GE(cycles, padTLLatency[i]))
+    padTLLatency[i] = cycles;
+  else
+    in ^= 0x10; // TL
+
   return (in & ~ctrl_reg) | (data_reg & ctrl_reg);
 }
 
+// update ports
+void PicoPortUpdate(void)
+{
+  if (port_lightgun || (Pico.m.hardware & PMS_HW_LG)) {
+    PicoIn.mouseInt[0] += PicoIn.mouse[0] - PicoIn.mouseInt[2];
+    PicoIn.mouseInt[1] += PicoIn.mouse[1] - PicoIn.mouseInt[3];
+    PicoIn.mouseInt[2] = PicoIn.mouse[0];
+    PicoIn.mouseInt[3] = PicoIn.mouse[1];
+    if (PicoIn.mouseInt[0] < 0) PicoIn.mouseInt[0] = 0;
+    if (PicoIn.mouseInt[0] > 320 ) PicoIn.mouseInt[0] = 320;
+    if (PicoIn.mouseInt[1] < 0) PicoIn.mouseInt[1] = 0;
+    if (PicoIn.mouseInt[1] > 240) PicoIn.mouseInt[1] = 240;
+  }
+}
+
+void PicoPortTrigger(void)
+{
+  // TODO mx/my scaling for V28/V30, H32/H40, soft/hardscaling?
+  int mx = PicoIn.mouseInt[0] + PicoIn.gunx;
+  int my = PicoIn.mouseInt[1] + PicoIn.guny;
+
+  if (unlikely(Pico.m.scanline == my) && port_lightgun &&
+      ((PicoMem.ioports[4]|PicoMem.ioports[5]) & 0x80)) {
+    // there's a rather massive delay from VDP, TV, light sensor, infrared
+    // transmission, to the TH pin; it lasts well into the next line.
+    if (port_type[1] == PICO_INPUT_JUSTIFIER) {
+      // justifier should trigger only if gun enabled?
+      if (!(PicoMem.ioports[2] & 0x30))
+        PicoVideoTriggerTH(mx + 66, my);
+    } else
+      PicoVideoTriggerTH(mx + 136, my);
+  }
+}
+
+// pad export for J-Cart
+u32 PicoReadPad(int i, u32 out_bits)
+{
+  return read_pad_3btn(i, out_bits);
+}
+
 void PicoSetInputDevice(int port, enum input_device device)
 {
   port_read_func *func;
+  int is_lg = device == PICO_INPUT_LIGHT_GUN || device == PICO_INPUT_JUSTIFIER;
 
   if (port < 0 || port > 2)
     return;
 
-  if (port == 1 && port_readers[0] == read_pad_team)
-    func = read_nothing;
-
-  else switch (device) {
-  case PICO_INPUT_PAD_3BTN:
-    func = read_pad_3btn;
-    break;
-
-  case PICO_INPUT_PAD_6BTN:
-    func = read_pad_6btn;
-    break;
-
-  case PICO_INPUT_PAD_TEAM:
-    func = read_pad_team;
-    break;
-
-  case PICO_INPUT_PAD_4WAY:
-    func = read_pad_4way;
-    break;
-
-  default:
-    func = read_nothing;
-    break;
+  switch (device) {
+  case PICO_INPUT_PAD_3BTN:   func = read_pad_3btn; break;
+  case PICO_INPUT_PAD_6BTN:   func = read_pad_6btn; break;
+  case PICO_INPUT_PAD_TEAM:   func = read_pad_team; break;
+  case PICO_INPUT_PAD_4WAY:   func = read_pad_4way; break;
+  case PICO_INPUT_MOUSE:      func = read_pad_mouse; break;
+  case PICO_INPUT_LIGHT_GUN:  func = read_pad_menacer; break;
+  case PICO_INPUT_JUSTIFIER:  func = read_pad_justifier; break;
+  default:                    func = read_nothing; break;
   }
 
+  if (port == 1 && port_type[0] == PICO_INPUT_PAD_TEAM)
+    func = read_nothing;
+
+  port_lightgun &= ~(1<<port);
+  port_lightgun |= is_lg<<port;
+  port_type[port] = device;
   port_readers[port] = func;
+
+  padTLLatency[port] = padTHLatency[port] = padTHTimeout[port] = SekCyclesDone();
 }
 
 NOINLINE u32 io_ports_read(u32 a)
@@ -486,25 +631,41 @@ NOINLINE void io_ports_write(u32 a, u32 d)
 {
   a = (a>>1) & 0xf;
 
-  // 6 button gamepad: if TH went from 0 to 1, gamepad changes state
+  // for some controllers, changing TH/TR changes controller state
   if (1 <= a && a <= 2)
   {
-    Pico.m.padDelay[a - 1] = 0;
-    if (port_readers[a - 1] == read_pad_team) {
+    // there's a timeout for toggling the TH line
+    if (CYCLES_GE(SekCyclesDone(), padTHTimeout[a - 1]))
+      Pico.m.padTHPhase[a - 1] = 0;
+    padTHTimeout[a - 1] = SekCyclesDone() + 25*488;
+
+    if (port_type[a - 1] == PICO_INPUT_PAD_TEAM) {
       if (d & 0x40)
         Pico.m.padTHPhase[a - 1] = 0;
       else if ((d^PicoMem.ioports[a]) & 0x60)
         Pico.m.padTHPhase[a - 1]++;
-    } else if (port_readers[0] == read_pad_4way) {
+    } else if (port_type[0] == PICO_INPUT_PAD_4WAY) {
       if (a == 2 && ((PicoMem.ioports[a] ^ d) & 0x70))
         Pico.m.padTHPhase[0] = 0;
       if (a == 1 && !(PicoMem.ioports[a] & 0x40) && (d & 0x40))
         Pico.m.padTHPhase[0]++;
+    } else if (port_type[a - 1] == PICO_INPUT_MOUSE) {
+      if ((d^PicoMem.ioports[a]) & 0x20) {
+        if (Pico.m.padTHPhase[a - 1]) { // in readout?
+          padTLLatency[a - 1] = SekCyclesDone() + 100;
+          Pico.m.padTHPhase[a - 1]++;
+        } else // in ID cycle
+          padTLLatency[a - 1] = SekCyclesDone() + 25; // Cannon Fodder
+      }
+      if ((d^PicoMem.ioports[a]) & 0x40) {
+        // 1->0 transition starts the readout protocol
+        Pico.m.padTHPhase[a - 1] = !(d & 0x40);
+      }
     } else if (!(PicoMem.ioports[a] & 0x40) && (d & 0x40))
       Pico.m.padTHPhase[a - 1]++;
   }
 
-  // after switching TH to input there's a latency before the pullup value is 
+  // after switching TH to input there's a latency before the pullup value is
   // read back as input (see Decap Attack, not in Samurai Showdown, 32x WWF Raw)
   if (4 <= a && a <= 5) {
     if ((PicoMem.ioports[a] & 0x40) && !(d & 0x40) && !(PicoMem.ioports[a - 3] & 0x40))
@@ -516,6 +677,64 @@ NOINLINE void io_ports_write(u32 a, u32 d)
   PicoMem.ioports[a] = d;
 }
 
+void io_ports_reset(void)
+{
+  int port;
+
+  for (port = 0; port < 3; port++)
+    padTLLatency[port] = padTHLatency[port] = padTHTimeout[port] = SekCyclesDone();
+}
+
+int io_ports_pack(void *buf, size_t size)
+{
+  size_t b, i;
+  memcpy(buf, PicoMem.ioports, (b = sizeof(PicoMem.ioports)));
+  for (i = 0; i < ARRAY_SIZE(Pico.m.padTHPhase); i++)
+    save_u8_(buf, &b, Pico.m.padTHPhase[i]);
+  for (i = 0; i < ARRAY_SIZE(Pico.m.padDelay); i++)
+    save_u8_(buf, &b, Pico.m.padDelay[i]);
+  for (i = 0; i < ARRAY_SIZE(padTHLatency); i++) {
+    save_s32(buf, &b, padTHLatency[i]);
+    save_s32(buf, &b, padTLLatency[i]);
+  }
+  for (i = 0; i < 4; i++)
+    save_u16(buf, &b, PicoIn.padInt[i]);
+  for (i = 0; i < 4; i++)
+    save_s32(buf, &b, PicoIn.mouseInt[i]);
+  for (i = 0; i < ARRAY_SIZE(padTHTimeout); i++)
+    save_u32(buf, &b, padTHTimeout[i]);
+  assert(b <= size);
+  return b;
+}
+
+void io_ports_unpack(const void *buf, size_t size)
+{
+  size_t b, i;
+  memcpy(PicoMem.ioports, buf, (b = sizeof(PicoMem.ioports)));
+  for (i = 0; i < ARRAY_SIZE(Pico.m.padTHPhase); i++)
+    Pico.m.padTHPhase[i] = load_u8_(buf, &b);
+  for (i = 0; i < ARRAY_SIZE(Pico.m.padDelay); i++)
+    Pico.m.padDelay[i] = load_u8_(buf, &b);
+  for (i = 0; i < ARRAY_SIZE(padTHLatency); i++) {
+    padTHLatency[i] = load_s32(buf, &b);
+    padTLLatency[i] = load_s32(buf, &b);
+  }
+  for (i = 0; i < 4; i++)
+    PicoIn.padInt[i] = load_u16(buf, &b);
+  for (i = 0; i < 4; i++)
+    PicoIn.mouseInt[i] = load_s32(buf, &b);
+
+  // convert pad delay from line count to samples
+  for (i = 0; i < ARRAY_SIZE(padTHTimeout); i++) {
+    if (b < size)
+      padTHTimeout[i] = load_u32(buf, &b);
+    else
+      padTHTimeout[i] = SekCyclesDone() - (25-Pico.m.padDelay[i]) * 488;
+  }
+
+  assert(b <= size);
+}
+
 static int z80_cycles_from_68k(void)
 {
   int m68k_cnt = SekCyclesDone() - Pico.t.m68c_frame_start;
@@ -530,20 +749,22 @@ void NOINLINE ctl_write_z80busreq(u32 d)
   {
     if (d)
     {
-      Pico.t.z80c_cnt = z80_cycles_from_68k() + (Pico.t.z80_busdelay >> 8);
+      Pico.t.z80c_aim = Pico.t.z80c_cnt = z80_cycles_from_68k() + 2;
+      Pico.t.z80c_cnt += Pico.t.z80_busdelay >> 8;
       Pico.t.z80_busdelay &= 0xff;
     }
     else
     {
       if ((PicoIn.opt & POPT_EN_Z80) && !Pico.m.z80_reset) {
-        // Z80 grants bus 2 cycles after the next M cycle, even within an insn
+        // Z80 grants bus after the current M cycle, even within an insn
         // simulate this by accumulating the last insn overhang in busdelay
-        unsigned granted = z80_cycles_from_68k() + 6;
+        unsigned granted;
         pprof_start(m68k);
         PicoSyncZ80(SekCyclesDone());
+        pprof_end_sub(m68k);
+        granted = Pico.t.z80c_aim + 6; // M cycle is 3-6 cycles
         Pico.t.z80_busdelay += (Pico.t.z80c_cnt - granted) << 8;
         Pico.t.z80c_cnt = granted;
-        pprof_end_sub(m68k);
       }
     }
     Pico.m.z80Run = d;
@@ -563,12 +784,13 @@ void NOINLINE ctl_write_z80reset(u32 d)
         PicoSyncZ80(SekCyclesDone());
         pprof_end_sub(m68k);
       }
+      Pico.t.z80_busdelay &= 0xff; // also resets bus request
       YM2612ResetChip();
       timers_reset();
     }
     else
     {
-      Pico.t.z80c_cnt = z80_cycles_from_68k() + 2;
+      Pico.t.z80c_aim = Pico.t.z80c_cnt = z80_cycles_from_68k() + 2;
       z80_reset();
     }
     Pico.m.z80_reset = d;
@@ -611,7 +833,7 @@ static u32 PicoRead8_sram(u32 a)
   // XXX: this is banking unfriendly
   if (a < Pico.romsize)
     return Pico.rom[MEM_BE2(a)];
-  
+ 
   return m68k_unmapped_read8(a);
 }
 
@@ -700,7 +922,7 @@ static u32 PicoRead8_z80(u32 a)
   if ((a & 0x4000) == 0x0000) {
     d = PicoMem.zram[a & 0x1fff];
   } else if ((a & 0x6000) == 0x4000) // 0x4000-0x5fff
-    d = ym2612_read_local_68k(); 
+    d = ym2612_read_local_68k();
   else {
     elprintf(EL_UIO|EL_ANOMALY, "68k bad read [%06x] @%06x", a, SekPc);
     d = (u8)PicoRead16_floating(a);
@@ -728,8 +950,7 @@ static void PicoWrite8_z80(u32 a, u32 d)
     return;
   }
   if ((a & 0x6000) == 0x4000) { // FM Sound
-    if (PicoIn.opt & POPT_EN_FM)
-      ym2612_write_local(a & 3, d & 0xff, 0);
+    ym2612_write_local(a & 3, d & 0xff, 0);
     return;
   }
   // TODO: probably other VDP access too? Maybe more mirrors?
@@ -767,9 +988,9 @@ u32 PicoRead8_io(u32 a)
     goto end;
   }
 
-  d = PicoRead16_floating(a);
-
   if ((a & 0xfc00) == 0x1000) {
+    d = (u8)PicoRead16_floating(a);
+
     if ((a & 0xff01) == 0x1100) { // z80 busreq (verified)
       // bit8 seems to be readable in this range
       if (!(a & 1)) {
@@ -799,10 +1020,10 @@ u32 PicoRead16_io(u32 a)
     goto end;
   }
 
-  d = PicoRead16_floating(a);
-
   // bit8 seems to be readable in this range
   if ((a & 0xfc00) == 0x1000) {
+    d = PicoRead16_floating(a);
+
     if ((a & 0xff00) == 0x1100) { // z80 busreq
       d &= ~0x0100;
       d |= (z80_cycles_from_68k() < Pico.t.z80c_cnt) << 8;
@@ -950,8 +1171,10 @@ PICO_INTERNAL void PicoMemSetup(void)
   mask = (1 << M68K_MEM_SHIFT) - 1;
   rs = (Pico.romsize + mask) & ~mask;
   if (rs > 0xa00000) rs = 0xa00000; // max cartridge area
-  cpu68k_map_set(m68k_read8_map,  0x000000, rs - 1, Pico.rom, 0);
-  cpu68k_map_set(m68k_read16_map, 0x000000, rs - 1, Pico.rom, 0);
+  if (rs) {
+    cpu68k_map_set(m68k_read8_map,  0x000000, rs - 1, Pico.rom, 0);
+    cpu68k_map_set(m68k_read16_map, 0x000000, rs - 1, Pico.rom, 0);
+  }
 
   // Common case of on-cart (save) RAM, usually at 0x200000-...
   if ((Pico.sv.flags & SRF_ENABLED) && Pico.sv.data != NULL) {
@@ -1113,7 +1336,7 @@ void ym2612_sync_timers(int z80_cycles, int mode_old, int mode_new)
        * Starting a timer takes place at the next tick, so xcycles needs to be
        * rounded up to that: t = next tick# = (xcycles / TICK_ZCYCLES) + 1
        */
-      unsigned t = ((xcycles * (((1LL<<32)/TIMER_A_TICK_ZCYCLES)+1))>>32) + 1;
+      unsigned t = DIVQ32(xcycles, TIMER_A_TICK_ZCYCLES) + 1;
       Pico.t.timer_a_next_oflow = t*TIMER_A_TICK_ZCYCLES + Pico.t.timer_a_step;
     }
   }
@@ -1136,7 +1359,7 @@ void ym2612_sync_timers(int z80_cycles, int mode_old, int mode_new)
        * reset by loading timer b. The first run of timer b after loading is
        * therefore shorter by up to 15 ticks.
        */
-      unsigned t = ((xcycles * (((1LL<<32)/TIMER_A_TICK_ZCYCLES)+1))>>32) + 1;
+      unsigned t = DIVQ32(xcycles, TIMER_A_TICK_ZCYCLES) + 1;
       int step = Pico.t.timer_b_step - TIMER_A_TICK_ZCYCLES*(t&15);
       Pico.t.timer_b_next_oflow = t*TIMER_A_TICK_ZCYCLES + step;
     }
@@ -1171,7 +1394,7 @@ static int ym2612_write_local(u32 a, u32 d, int is_from_z80)
 
       // the busy flag in the YM2612 status is actually a 32 cycle timer
       // (89.6 Z80 cycles), triggered by any write to the data port.
-      Pico.t.ym2612_busy = (cycles + 90) << 8; // Q8 for convenience
+      Pico.t.ym2612_busy = (cycles << 8) + YMBUSY_ZCYCLES; // Q8 for convenience
 
       switch (addr)
       {
@@ -1231,7 +1454,7 @@ static int ym2612_write_local(u32 a, u32 d, int is_from_z80)
           //elprintf(EL_STATUS, "%03i dac w %08x z80 %i", cycles, d, is_from_z80);
           if (ym2612.dacen)
             PsndDoDAC(cycles);
-          ym2612.dacout = ((int)d - 0x80) << 6;
+          ym2612.dacout = ((int)d - 0x80) << DAC_SHIFT;
           return 0;
         }
         case 0x2b: { /* DAC Sel  (YM2612) */
@@ -1278,33 +1501,60 @@ static u32 ym2612_read_local_68k(void)
   return ym2612.OPN.ST.status;
 }
 
-void ym2612_pack_state(void)
+// legacy code, only used for GP2X
+void ym2612_pack_state_old(void)
 {
   // timers are saved as tick counts, in 16.16 int format
-  int tac, tat = 0, tbc, tbt = 0;
-  tac = 1024 - ym2612.OPN.ST.TA;
-  tbc = 256  - ym2612.OPN.ST.TB;
-  if (Pico.t.timer_a_next_oflow != TIMER_NO_OFLOW)
-    tat = (int)((double)(Pico.t.timer_a_step - Pico.t.timer_a_next_oflow)
-          / (double)Pico.t.timer_a_step * tac * 65536);
-  if (Pico.t.timer_b_next_oflow != TIMER_NO_OFLOW)
-    tbt = (int)((double)(Pico.t.timer_b_step - Pico.t.timer_b_next_oflow)
-          / (double)Pico.t.timer_b_step * tbc * 65536);
-  elprintf(EL_YMTIMER, "save: timer a %i/%i", tat >> 16, tac);
-  elprintf(EL_YMTIMER, "save: timer b %i/%i", tbt >> 16, tbc);
+  int tat = 0, tbt = 0, busy = 0;
+  u8 buf[16];
+  size_t b = 0;
+
+  ym2612_pack_timers(buf, sizeof(buf));
+  load_u16(buf, &b), load_u16(buf, &b);
+  tat = load_u32(buf, &b);
+  tbt = load_u32(buf, &b);
+  busy = load_u32(buf, &b);
 
 #ifdef __GP2X__
   if (PicoIn.opt & POPT_EXT_FM)
     YM2612PicoStateSave2_940(tat, tbt);
   else
 #endif
-    YM2612PicoStateSave2(tat, tbt);
+    YM2612PicoStateSave2(tat, tbt, busy);
 }
 
-void ym2612_unpack_state(void)
+int ym2612_pack_timers(void *buf, size_t size)
 {
-  int i, ret, tac, tat, tbc, tbt;
+  // timers are saved as tick counts, in 16.16 int format
+  int tac, tat = 0, tbc, tbt = 0, busy = 0;
+  size_t b = 0;
+
+  tac = 1024 - ym2612.OPN.ST.TA;
+  tbc = 256  - ym2612.OPN.ST.TB;
+  if (Pico.t.ym2612_busy > 0)
+    busy = cycles_z80_to_68k(Pico.t.ym2612_busy);
+  if (Pico.t.timer_a_next_oflow != TIMER_NO_OFLOW)
+    tat = MULQ((Pico.t.timer_a_step - Pico.t.timer_a_next_oflow), INVQ32(TIMER_A_TICK_ZCYCLES), 16);
+  if (Pico.t.timer_b_next_oflow != TIMER_NO_OFLOW)
+    tbt = MULQ((Pico.t.timer_b_step - Pico.t.timer_b_next_oflow), INVQ32(TIMER_B_TICK_ZCYCLES), 16);
+  elprintf(EL_YMTIMER, "save: timer a %i/%i", tat >> 16, tac);
+  elprintf(EL_YMTIMER, "save: timer b %i/%i", tbt >> 16, tbc);
+
+  assert(size >= 16);
+  save_u16(buf, &b, ym2612.OPN.ST.TA);
+  save_u16(buf, &b, ym2612.OPN.ST.TB);
+  save_u32(buf, &b, tat);
+  save_u32(buf, &b, tbt);
+  save_u32(buf, &b, busy);
+  return b;
+}
+
+// legacy code, only used for GP2X
+void ym2612_unpack_state_old(void)
+{
+  int i, ret, tat, tbt, busy = 0;
   YM2612PicoStateLoad();
+  Pico.t.m68c_frame_start = SekCyclesDone();
 
   // feed all the registers and update internal state
   for (i = 0x20; i < 0xA0; i++) {
@@ -1333,20 +1583,44 @@ void ym2612_unpack_state(void)
     ret = YM2612PicoStateLoad2_940(&tat, &tbt);
   else
 #endif
-    ret = YM2612PicoStateLoad2(&tat, &tbt);
+    ret = YM2612PicoStateLoad2(&tat, &tbt, &busy);
   if (ret != 0) {
     elprintf(EL_STATUS, "old ym2612 state");
     return; // no saved timers
   }
+  {
+    u8 tmp[16];
+    size_t b = 0;
+    save_u16(tmp, &b, ym2612.OPN.ST.TA);
+    save_u16(tmp, &b, ym2612.OPN.ST.TB);
+    save_u32(tmp, &b, tat);
+    save_u32(tmp, &b, tbt);
+    save_u32(tmp, &b, busy);
+    ym2612_unpack_timers(tmp, b);
+  }
+}
 
+void ym2612_unpack_timers(const void *buf, size_t size)
+{
+  int tac, tat, tbc, tbt, busy;
+  size_t b = 0;
+  assert(size >= 16);
+  if (size < 16)
+    return;
+  ym2612.OPN.ST.TA = load_u16(buf, &b);
+  ym2612.OPN.ST.TB = load_u16(buf, &b);
+  tat = load_u32(buf, &b);
+  tbt = load_u32(buf, &b);
+  busy = load_u32(buf, &b);
+  Pico.t.ym2612_busy = cycles_68k_to_z80(busy);
   tac = (1024 - ym2612.OPN.ST.TA) << 16;
   tbc = (256  - ym2612.OPN.ST.TB) << 16;
   if (ym2612.OPN.ST.mode & 1)
-    Pico.t.timer_a_next_oflow = (int)((double)(tac - tat) / (double)tac * Pico.t.timer_a_step);
+    Pico.t.timer_a_next_oflow = (1LL * (tac-tat) * TIMER_A_TICK_ZCYCLES)>>16;
   else
     Pico.t.timer_a_next_oflow = TIMER_NO_OFLOW;
   if (ym2612.OPN.ST.mode & 2)
-    Pico.t.timer_b_next_oflow = (int)((double)(tbc - tbt) / (double)tbc * Pico.t.timer_b_step);
+    Pico.t.timer_b_next_oflow = (1LL * (tbc-tbt) * TIMER_B_TICK_ZCYCLES)>>16;
   else
     Pico.t.timer_b_next_oflow = TIMER_NO_OFLOW;
   elprintf(EL_YMTIMER, "load: %i/%i, timer_a_next_oflow %i", tat>>16, tac>>16, Pico.t.timer_a_next_oflow >> 8);
@@ -1367,8 +1641,6 @@ void PicoWrite16_32x(u32 a, u32 d) {}
 static void access_68k_bus(int delay) // bus delay as Q8
 {
   // TODO: if the 68K is in DMA wait, Z80 has to wait until DMA ends
-  if (Pico.video.status & (PVS_CPUWR|PVS_CPURD))
-    z80_subCLeft(z80_cyclesLeft); // rather rough on both condition and action
 
   // 68k bus access delay for z80. The fractional part needs to be accumulated
   // until an additional cycle is full. That is then added to the integer part.
@@ -1377,7 +1649,7 @@ static void access_68k_bus(int delay) // bus delay as Q8
   Pico.t.z80_busdelay &= 0xff; // leftover cycle fraction
   // don't use SekCyclesBurn() here since the Z80 doesn't run in cycle lock to
   // the 68K. Count the stolen cycles to be accounted later in the 68k CPU runs
-  Pico.t.z80_buscycles += 8;
+  Pico.t.z80_buscycles += 0x80; // TODO <=8.4 for Rick 2, but >=8.9 for misc_test
 }
 
 static unsigned char z80_md_vdp_read(unsigned short a)
@@ -1423,8 +1695,7 @@ static unsigned char z80_md_bank_read(unsigned short a)
 
 static void z80_md_ym2612_write(unsigned int a, unsigned char data)
 {
-  if (PicoIn.opt & POPT_EN_FM)
-    ym2612_write_local(a, data, 1);
+  ym2612_write_local(a, data, 1);
 }
 
 static void z80_md_vdp_br_write(unsigned int a, unsigned char data)
