@@ -2,6 +2,7 @@
  * PicoDrive
  * (c) Copyright Dave, 2004
  * (C) notaz, 2006-2009
+ * (C) irixxxx, 2019-2024
  *
  * This work is licensed under the terms of MAME license.
  * See COPYING file in the top-level directory.
@@ -10,11 +11,14 @@
 #include <string.h>
 #include "../pico_int.h"
 #include "ym2612.h"
+#include "ym2413.h"
 #include "sn76496.h"
-#include "emu2413/emu2413.h"
+#include "../cd/megasd.h"
 #include "resampler.h"
 #include "mix.h"
 
+#define YM2612_CH6PAN   0x1b6   // panning register for channel 6 (used for DAC)
+
 void (*PsndMix_32_to_16)(s16 *dest, s32 *src, int count) = mix_32_to_16_stereo;
 
 // master int buffer to mix to
@@ -24,47 +28,10 @@ static s32 PsndBuffer[2*(54000+100)/50+2];
 // cdda output buffer
 s16 cdda_out_buffer[2*1152];
 
-// sn76496
-extern int *sn76496_regs;
-
 // FM resampling polyphase FIR
 static resampler_t *fmresampler;
 static int (*PsndFMUpdate)(s32 *buffer, int length, int stereo, int is_buf_empty);
 
-// ym2413
-static OPLL *opll = NULL;
-static OPLL old_opll;
-static struct {
-  uint32_t adr;
-  uint8_t reg[sizeof(opll->reg)];
-} opll_buf;
-
-
-void YM2413_regWrite(unsigned data){
-  OPLL_writeIO(opll,0,data);
-}
-void YM2413_dataWrite(unsigned data){
-  OPLL_writeIO(opll,1,data);
-}
-
-PICO_INTERNAL void *YM2413GetRegs(void)
-{
-  memcpy(opll_buf.reg, opll->reg, sizeof(opll->reg));
-  opll_buf.adr = opll->adr;
-  return &opll_buf;
-}
-
-PICO_INTERNAL void YM2413UnpackState(void)
-{
-  int i;
-
-  for (i = sizeof(opll->reg)-1; i >= 0; i--) {
-    OPLL_writeIO(opll, 0, i);
-    OPLL_writeIO(opll, 1, opll_buf.reg[i]);
-  }
-  opll->adr = opll_buf.adr;
-}
-
 PICO_INTERNAL void PsndInit(void)
 {
   opll = OPLL_new(OSC_NTSC/15, OSC_NTSC/15/72);
@@ -88,7 +55,7 @@ PICO_INTERNAL void PsndReset(void)
 }
 
 // FM polyphase FIR resampling
-#define FMFIR_TAPS	9
+#define FMFIR_TAPS	8
 
 // resample FM from its native 53267Hz/52781Hz with polyphase FIR filter
 static int ymchans;
@@ -125,7 +92,7 @@ static void YMFM_setup_FIR(int inrate, int outrate, int stereo)
 {
   int mindiff = 999;
   int diff, mul, div;
-  int minmult = 22, maxmult = 55; // min,max interpolation factor
+  int minmult = 11, maxmult = 61; // min,max interpolation factor
 
   // compute filter ratio with largest multiplier for smallest error
   for (mul = minmult; mul <= maxmult; mul++) {
@@ -135,7 +102,7 @@ static void YMFM_setup_FIR(int inrate, int outrate, int stereo)
       mindiff = diff;
       Pico.snd.fm_fir_mul = mul;
       Pico.snd.fm_fir_div = div;
-      if (abs(mindiff) <= inrate/1000+1) break; // below error limit
+      if (abs(mindiff)*1000 <= inrate) break; // below error limit
     }
   }
   printf("FM polyphase FIR ratio=%d/%d error=%.3f%%\n",
@@ -152,6 +119,10 @@ static int YM2612UpdateONE(s32 *buffer, int length, int stereo, int is_buf_empty
   return YM2612UpdateOne(buffer, length, stereo, is_buf_empty);
 }
 
+static int ymclock;
+static int ymrate;
+static int ymopts;
+
 // to be called after changing sound rate or chips
 void PsndRerate(int preserve_state)
 {
@@ -159,54 +130,56 @@ void PsndRerate(int preserve_state)
   int target_fps = Pico.m.pal ? 50 : 60;
   int target_lines = Pico.m.pal ? 313 : 262;
   int sms_clock = Pico.m.pal ? OSC_PAL/15 : OSC_NTSC/15;
+  int ym2413_rate = (sms_clock + 36) / 72;
   int ym2612_clock = Pico.m.pal ? OSC_PAL/7 : OSC_NTSC/7;
   int ym2612_rate = YM2612_NATIVE_RATE();
-  int ym2413_rate = (sms_clock + 36) / 72;
+  int ym2612_init = !preserve_state;
+  int state_size = 4096;
 
-  if (preserve_state) {
-    state = malloc(0x204);
-    if (state == NULL) return;
-    ym2612_pack_state();
-    memcpy(state, YM2612GetRegs(), 0x204);
+  // don't init YM2612 if preserve_state and no parameter changes
+  ym2612_init |= ymclock != ym2612_clock || ymopts != (PicoIn.opt & (POPT_DIS_FM_SSGEG|POPT_EN_FM_DAC));
+  ym2612_init |= ymrate != (PicoIn.opt & POPT_EN_FM_FILTER ? ym2612_rate : PicoIn.sndRate);
+  ymclock = ym2612_clock;
+  ymrate = (PicoIn.opt & POPT_EN_FM_FILTER ? ym2612_rate : PicoIn.sndRate);
+  ymopts = PicoIn.opt & (POPT_DIS_FM_SSGEG|POPT_EN_FM_DAC);
 
-    if (opll != NULL)
-       memcpy(&old_opll, opll, sizeof(OPLL)); // remember old state
+  if (preserve_state && ym2612_init) {
+    state = malloc(state_size);
+    if (state)
+      state_size = YM2612PicoStateSave3(state, state_size);
   }
+
   if (PicoIn.AHW & PAHW_SMS) {
     OPLL_setRate(opll, ym2413_rate);
-    OPLL_reset(opll);
+    if (!preserve_state)
+      OPLL_reset(opll);
     YMFM_setup_FIR(ym2413_rate, PicoIn.sndRate, 0);
     PsndFMUpdate = YM2413UpdateFIR;
   } else if ((PicoIn.opt & POPT_EN_FM_FILTER) && ym2612_rate != PicoIn.sndRate) {
     // polyphase FIR resampler, resampling directly from native to output rate
-    YM2612Init(ym2612_clock, ym2612_rate,
+    if (ym2612_init)
+      YM2612Init(ym2612_clock, ym2612_rate,
         ((PicoIn.opt&POPT_DIS_FM_SSGEG) ? 0 : ST_SSG) |
         ((PicoIn.opt&POPT_EN_FM_DAC)    ? ST_DAC : 0));
     YMFM_setup_FIR(ym2612_rate, PicoIn.sndRate, PicoIn.opt & POPT_EN_STEREO);
     PsndFMUpdate = YM2612UpdateFIR;
   } else {
-    YM2612Init(ym2612_clock, PicoIn.sndRate,
+    if (ym2612_init)
+      YM2612Init(ym2612_clock, PicoIn.sndRate,
         ((PicoIn.opt&POPT_DIS_FM_SSGEG) ? 0 : ST_SSG) |
         ((PicoIn.opt&POPT_EN_FM_DAC)    ? ST_DAC : 0));
     PsndFMUpdate = YM2612UpdateONE;
   }
-  if (preserve_state) {
-    // feed it back it's own registers, just like after loading state
-    memcpy(YM2612GetRegs(), state, 0x204);
-    ym2612_unpack_state();
 
-    if (opll != NULL) {
-      memcpy(&opll->adr, &old_opll.adr, sizeof(OPLL)-20); // restore old state
-      OPLL_forceRefresh(opll);
-    }
+  if (state) {
+    YM2612PicoStateLoad3(state, state_size);
+    free(state);
   }
 
-  if (preserve_state) memcpy(state, sn76496_regs, 28*4); // remember old state
-  SN76496_init(Pico.m.pal ? OSC_PAL/15 : OSC_NTSC/15, PicoIn.sndRate);
-  if (preserve_state) memcpy(sn76496_regs, state, 28*4); // restore old state
-
-  if (state)
-    free(state);
+  if (preserve_state)
+    SN76496_set_clockrate(Pico.m.pal ? OSC_PAL/15 : OSC_NTSC/15, PicoIn.sndRate);
+  else
+    SN76496_init(Pico.m.pal ? OSC_PAL/15 : OSC_NTSC/15, PicoIn.sndRate);
 
   // calculate Pico.snd.len
   Pico.snd.len = PicoIn.sndRate / target_fps;
@@ -216,8 +189,8 @@ void PsndRerate(int preserve_state)
   // samples per line (Q16)
   Pico.snd.smpl_mult = 65536LL * PicoIn.sndRate / (target_fps*target_lines);
   // samples per z80 clock (Q20)
-  Pico.snd.clkl_mult = 16 * Pico.snd.smpl_mult * 15/7 / 488.5;
-  // samples per 44.1 KHz sample
+  Pico.snd.clkz_mult = 16 * Pico.snd.smpl_mult * 15/7 / 488.5;
+  // samples per 44.1 KHz sample (Q16)
   Pico.snd.cdda_mult = 65536LL * 44100 / PicoIn.sndRate;
   Pico.snd.cdda_div  = 65536LL * PicoIn.sndRate / 44100;
 
@@ -256,7 +229,7 @@ PICO_INTERNAL void PsndDoDAC(int cyc_to)
   if (!PicoIn.sndOut) return;
 
   // number of samples to fill in buffer (Q20)
-  len = (cyc_to * Pico.snd.clkl_mult) - Pico.snd.dac_pos;
+  len = (cyc_to * Pico.snd.clkz_mult) - Pico.snd.dac_pos;
 
   // update position and calculate buffer offset and length
   pos = (Pico.snd.dac_pos+0x80000) >> 20;
@@ -274,9 +247,12 @@ PICO_INTERNAL void PsndDoDAC(int cyc_to)
   // 1 sample delay for correct IIR filtering over audio frame boundaries
   if (PicoIn.opt & POPT_EN_STEREO) {
     s16 *d = PicoIn.sndOut + pos*2;
-    // left channel only, mixed ro right channel in mixing phase
-    *d++ += Pico.snd.dac_val2, *d++ += Pico.snd.dac_val2;
-    while (--len) *d++ += Pico.snd.dac_val, *d++ += Pico.snd.dac_val;
+    int pan = ym2612.REGS[YM2612_CH6PAN];
+    int l = pan & 0x80 ? Pico.snd.dac_val : 0;
+    int r = pan & 0x40 ? Pico.snd.dac_val : 0;
+    *d++ += pan & 0x80 ? Pico.snd.dac_val2 : 0;
+    *d++ += pan & 0x40 ? Pico.snd.dac_val2 : 0;
+    while (--len) *d++ += l, *d++ += r;
   } else {
     s16 *d = PicoIn.sndOut + pos;
     *d++ += Pico.snd.dac_val2;
@@ -295,7 +271,7 @@ PICO_INTERNAL void PsndDoPSG(int cyc_to)
   if (!PicoIn.sndOut) return;
 
   // number of samples to fill in buffer (Q20)
-  len = (cyc_to * Pico.snd.clkl_mult) - Pico.snd.psg_pos;
+  len = (cyc_to * Pico.snd.clkz_mult) - Pico.snd.psg_pos;
 
   // update position and calculate buffer offset and length
   pos = (Pico.snd.psg_pos+0x80000) >> 20;
@@ -325,7 +301,7 @@ PICO_INTERNAL void PsndDoSMSFM(int cyc_to)
   if (!PicoIn.sndOut) return;
 
   // number of samples to fill in buffer (Q20)
-  len = (cyc_to * Pico.snd.clkl_mult) - Pico.snd.ym2413_pos;
+  len = (cyc_to * Pico.snd.clkz_mult) - Pico.snd.ym2413_pos;
 
   // update position and calculate buffer offset and length
   pos = (Pico.snd.ym2413_pos+0x80000) >> 20;
@@ -366,7 +342,7 @@ PICO_INTERNAL void PsndDoFM(int cyc_to)
   if (!PicoIn.sndOut) return;
 
   // Q20, number of samples since last call
-  len = (cyc_to * Pico.snd.clkl_mult) - Pico.snd.fm_pos;
+  len = (cyc_to * Pico.snd.clkz_mult) - Pico.snd.fm_pos;
 
   // update position and calculate buffer offset and length
   pos = (Pico.snd.fm_pos+0x80000) >> 20;
@@ -393,7 +369,7 @@ PICO_INTERNAL void PsndDoPCM(int cyc_to)
   if (!PicoIn.sndOut) return;
 
   // Q20, number of samples since last call
-  len = (cyc_to * Pico.snd.clkl_mult) - Pico.snd.pcm_pos;
+  len = (cyc_to * Pico.snd.clkz_mult) - Pico.snd.pcm_pos;
 
   // update position and calculate buffer offset and length
   pos = (Pico.snd.pcm_pos+0x80000) >> 20;
@@ -415,13 +391,23 @@ static void cdda_raw_update(s32 *buffer, int length, int stereo)
 {
   int ret, cdda_bytes;
 
+  // apply start offset in frame (offset to 1st lba to play)
+  int offs = Pico_mcd->cdda_frame_offs * Pico.snd.cdda_div >> 16;
+  length -= offs;
+  buffer += offs * (stereo ? 2 : 1);
+  Pico_mcd->cdda_frame_offs = 0;
+
   cdda_bytes = (length * Pico.snd.cdda_mult >> 16) * 4;
 
+  // compute offset of last played sample in this frame (need for save/load)
+  Pico_mcd->m.cdda_lba_offset += cdda_bytes/4;
+  while (Pico_mcd->m.cdda_lba_offset >= 2352/4)
+    Pico_mcd->m.cdda_lba_offset -= 2352/4;
+
   ret = pm_read_audio(cdda_out_buffer, cdda_bytes, Pico_mcd->cdda_stream);
   if (ret < cdda_bytes) {
     memset((char *)cdda_out_buffer + ret, 0, cdda_bytes - ret);
     Pico_mcd->cdda_stream = NULL;
-    return;
   }
 
   // now mix
@@ -447,7 +433,9 @@ void cdda_start_play(int lba_base, int lba_offset, int lb_len)
     return;
   }
 
-  pm_seek(Pico_mcd->cdda_stream, (lba_base + lba_offset) * 2352, SEEK_SET);
+  // on restart after loading, consider offset of last played sample
+  pm_seek(Pico_mcd->cdda_stream, (lba_base + lba_offset) * 2352 +
+                                  Pico_mcd->m.cdda_lba_offset * 4, SEEK_SET);
   if (Pico_mcd->cdda_type == CT_WAV)
   {
     // skip headers, assume it's 44kHz stereo uncompressed
@@ -508,13 +496,21 @@ static int PsndRender(int offset, int length)
 
   // Fill up DAC output in case of missing samples (Q rounding errors)
   if (length-daclen > 0 && PicoIn.sndOut) {
-    s16 *dacbuf = PicoIn.sndOut + (daclen << stereo);
     Pico.snd.dac_pos += (length-daclen) << 20;
-    *dacbuf++ += Pico.snd.dac_val2;
-    if (stereo) *dacbuf++ += Pico.snd.dac_val2;
-    for (daclen++; length-daclen > 0; daclen++) {
-      *dacbuf++ += Pico.snd.dac_val;
-      if (stereo) *dacbuf++ += Pico.snd.dac_val;
+    if (PicoIn.opt & POPT_EN_STEREO) {
+      s16 *d = PicoIn.sndOut + daclen*2;
+      int pan = ym2612.REGS[YM2612_CH6PAN];
+      int l = pan & 0x80 ? Pico.snd.dac_val : 0;
+      int r = pan & 0x40 ? Pico.snd.dac_val : 0;
+      *d++ += pan & 0x80 ? Pico.snd.dac_val2 : 0;
+      *d++ += pan & 0x40 ? Pico.snd.dac_val2 : 0;
+      if (l|r) for (daclen++; length-daclen > 0; daclen++)
+          *d++ += l, *d++ += r;
+    } else {
+      s16 *d = PicoIn.sndOut + daclen;
+      *d++ += Pico.snd.dac_val2;
+      if (Pico.snd.dac_val) for (daclen++; length-daclen > 0; daclen++)
+        *d++ += Pico.snd.dac_val;
     }
     Pico.snd.dac_val2 = Pico.snd.dac_val;
   }
@@ -536,7 +532,7 @@ static int PsndRender(int offset, int length)
   // CD mode, cdda enabled, not data track, CDC is reading
   if ((PicoIn.AHW & PAHW_MCD) && (PicoIn.opt & POPT_EN_MCD_CDDA)
       && Pico_mcd->cdda_stream != NULL
-      && !(Pico_mcd->s68k_regs[0x36] & 1))
+      && (!(Pico_mcd->s68k_regs[0x36] & 1) || Pico_msd.state == 3))
   {
     if (Pico_mcd->cdda_type == CT_MP3)
       mp3_update(buf32, length-offset, stereo);
