@@ -2,7 +2,7 @@
  * PicoDrive
  * (c) Copyright Dave, 2004
  * (C) notaz, 2006-2009
- * (C) kub, 2020,2021
+ * (C) irixxxx, 2020-2024
  *
  * This work is licensed under the terms of MAME license.
  * See COPYING file in the top-level directory.
@@ -23,7 +23,8 @@ enum { clkdiv = 2 };    // CPU clock granularity: one of 1,2,4,8
 // Slot clock is sysclock/20 for h32 and sysclock/16 for h40.
 // One scanline is 63.7us/64.3us (ntsc/pal) long which is ~488.57 68k cycles.
 // Approximate by 488 for VDP.
-// 1 slot is 488/171 = 2.8538 68k cycles in h32, and 488/210 = 2.3238 in h40.
+// 1 slot is 20/7 = 2.857 68k cycles in h32, and 16/7 = 2.286 in h40. That's
+// 171 slots in h32, and ~214 (really 193 plus 17 prolonged in HSYNC) in h40.
 enum { slcpu = 488 };
 
 // VDP has a slot counter running from 0x00 to 0xff every scanline, but it has
@@ -34,6 +35,11 @@ enum { slcpu = 488 };
 enum { hint32 = 0x85, gapstart32 = 0x94, gapend32 = 0xe9};
 enum { hint40 = 0xa5, gapstart40 = 0xb7, gapend40 = 0xe5};
 
+// Basic timing in h32: 38 slots (~108.5 cycles) from hint to VDP output start
+// at slot 0x00. vint takes place on the 1st VBLANK line in slot 0x01 (~111.5).
+// Rendering takes 128 slots (~365.5), and right border starts at slot 0x80
+// (~474 cycles). hint occurs after 5 slots into the border (~488.5 cycles).
+
 // The horizontal sync period (HBLANK) is 30/37 slots (h32/h40):
 // h32: 4 slots front porch (1.49us), 13 HSYNC (4.84us), 13 back porch (4.84us)
 // h40: 5 slots front porch (1.49us), 16 HSYNC (4.77us), 16 back porch (4.77us)
@@ -183,7 +189,7 @@ int (*PicoDmaHook)(u32 source, int len, unsigned short **base, u32 *mask) = NULL
  */
 
 // NB code assumes fifo_* arrays have size 2^n
-static struct VdpFIFO { // XXX this must go into save file!
+static struct VdpFIFO {
   // last transferred FIFO data, ...x = index  XXX currently only CPU
   u16 fifo_data[4], fifo_dx;
 
@@ -399,7 +405,7 @@ int PicoVideoFIFOHint(void)
   // only need to refresh sprite position if we are synced
   if (Pico.est.DrawScanline == Pico.m.scanline && !(pv->status & SR_VB))
     PicoDrawRefreshSprites();
- 
+
   // if CPU is waiting for the bus, advance CPU and FIFO until bus is free
   if (pv->status & PVS_CPUWR)
     burn = PicoVideoFIFODrain(4, lc, 0);
@@ -426,6 +432,8 @@ void PicoVideoFIFOMode(int active, int h40)
 
   if (vf->fifo_maxslot)
     PicoVideoFIFOSync(lc);
+  else
+    lc = 0;
 
   vf->fifo_cyc2sl = vdpcyc2sl[active][h40];
   vf->fifo_sl2cyc = vdpsl2cyc[active][h40];
@@ -534,6 +542,9 @@ static void DmaSlow(int len, u32 source)
 
   SekCyclesBurnRun(PicoVideoFIFOWrite(len, FQ_FGDMA | (pvid->type == 1),
                               PVS_DMABG, SR_DMA | PVS_CPUWR));
+  // short transfers might have been completely conveyed to FIFO, adjust state
+  if ((pvid->status & SR_DMA) && VdpFIFO.fifo_total <= 4)
+    SetFIFOState(&VdpFIFO, pvid);
 
   if ((source & 0xe00000) == 0xe00000) { // Ram
     base = (u16 *)PicoMem.ram;
@@ -543,8 +554,8 @@ static void DmaSlow(int len, u32 source)
   {
     u8 r3 = Pico_mcd->s68k_regs[3];
     elprintf(EL_VDPDMA, "DmaSlow CD, r3=%02x", r3);
-    if (source < Pico.romsize /*0x20000*/) { // Bios area
-      base = (u16 *)(Pico.rom + (source & 0xfe0000));
+    if ((source & 0xfc0000) == pcd_base_address) { // Bios area
+      base = (u16 *)(Pico_mcd->bios + (source & 0xfe0000));
     } else if ((source & 0xfc0000) == pcd_base_address+0x200000) { // Word Ram
       if (!(r3 & 4)) { // 2M mode
         base = (u16 *)(Pico_mcd->word_ram2M + (source & 0x20000));
@@ -561,7 +572,8 @@ static void DmaSlow(int len, u32 source)
     } else if ((source & 0xfe0000) == pcd_base_address+0x020000) { // Prg Ram
       base = (u16 *)Pico_mcd->prg_ram_b[r3 >> 6];
       source -= 2; // XXX: test
-    }
+    } else // Rom
+      base = m68k_dma_source(source);
   }
   else
   {
@@ -591,7 +603,7 @@ static void DmaSlow(int len, u32 source)
           !((source ^ (source + len-1)) & ~mask))
       {
         // most used DMA mode
-        memcpy((char *)r + a, base + (source & mask), len * 2);
+        memcpy((char *)r + (u16)a, base + (source & mask), len * 2);
         a += len * 2;
         break;
       }
@@ -615,7 +627,8 @@ static void DmaSlow(int len, u32 source)
         if (sl > VdpFIFO.fifo_hcounts[0]-5) // hint delay is 5 slots
           sl = (s8)sl;
         // TODO this is needed to cover timing inaccuracies
-        if (sl <= 12)  sl = -2;
+        if (sl <= 12) sl = -3;
+        else if (sl <= 40) sl = 30;
         PicoDrawBgcDMA(base, source, mask, len, sl);
         // do last DMA cycle since it's all going to the same cram location
         source = source+len-1;
@@ -893,8 +906,8 @@ PICO_INTERNAL_ASM void PicoVideoWrite(u32 a,unsigned short d)
       // the vertical scroll value for this line must be read from VSRAM early,
       // since the A/B tile row to be read depends on it. E.g. Skitchin, OD2
       // in contrast, CRAM writes would have an immediate effect on the current
-      // pixel. XXX think about different offset values for different RAM types
-      PicoVideoSync(InHblank(30));
+      // pixel, so sync can be closer to start of actual image data
+      PicoVideoSync(InHblank(pvid->type == 3 ? 103 : 30)); // cram in Toy Story
 
     if (!(PicoIn.opt&POPT_DIS_VDP_FIFO))
     {
@@ -944,19 +957,16 @@ PICO_INTERNAL_ASM void PicoVideoWrite(u32 a,unsigned short d)
         }
 
         d &= 0xff;
-        if (num == 0 && !(pvid->reg[0]&2) && (d&2))
-          pvid->hv_latch = PicoVideoRead(0x08);
-        if (num == 12 && ((pvid->reg[12]^d)&0x01))
-          PicoVideoFIFOMode(pvid->reg[1]&0x40, d & 1);
 
         if (num == 1 && ((pvid->reg[1]^d)&0x40)) {
-          PicoVideoFIFOMode(d & 0x40, pvid->reg[12]&1);
           // handle line blanking before line rendering. Only the last switch
-          // before the 1st sync for other reasons is honoured.
-          PicoVideoSync(1);
-          lineenabled = (d&0x40) ? Pico.m.scanline : -1;
-          linedisabled = (d&0x40) ? -1 : Pico.m.scanline;
-          lineoffset = SekCyclesDone() - Pico.t.m68c_line_start;
+          // before the 1st sync for other reasons is honoured. Switching after
+          // active area is on next line
+          int skip = InHblank(470); // Deadly Moves
+          PicoVideoSync(skip);
+          lineenabled = (d&0x40) ? Pico.m.scanline + !skip: -1;
+          linedisabled = (d&0x40) ? -1 : Pico.m.scanline + !skip;
+          lineoffset = (skip ? SekCyclesDone() - Pico.t.m68c_line_start : 0);
         } else if (((1<<num) & 0x738ff) && pvid->reg[num] != d)
           // VDP regs 0-7,11-13,16-18 influence rendering, ignore all others
           PicoVideoSync(InHblank(93)); // Toy Story
@@ -965,15 +975,21 @@ PICO_INTERNAL_ASM void PicoVideoWrite(u32 a,unsigned short d)
         switch (num)
         {
           case 0x00:
+            if ((~dold&d)&2) {
+              unsigned c = SekCyclesDone() - Pico.t.m68c_line_start;
+              pvid->hv_latch = VdpFIFO.fifo_hcounts[c/clkdiv] | (pvid->v_counter << 8);
+            }
             elprintf(EL_INTSW, "hint_onoff: %i->%i [%u] pend=%i @ %06x", (dold&0x10)>>4,
                     (d&0x10)>>4, SekCyclesDone(), (pvid->pending_ints&0x10)>>4, SekPc);
             goto update_irq;
           case 0x01:
-            elprintf(EL_INTSW, "vint_onoff: %i->%i [%u] pend=%i @ %06x", (dold&0x20)>>5,
-                    (d&0x20)>>5, SekCyclesDone(), (pvid->pending_ints&0x20)>>5, SekPc);
+            if ((d^dold)&0x40)
+              PicoVideoFIFOMode(d & 0x40, pvid->reg[12]&1);
             if (!(pvid->status & PVS_VB2))
               pvid->status &= ~SR_VB;
             pvid->status |= ((d >> 3) ^ SR_VB) & SR_VB; // forced blanking
+            elprintf(EL_INTSW, "vint_onoff: %i->%i [%u] pend=%i @ %06x", (dold&0x20)>>5,
+                    (d&0x20)>>5, SekCyclesDone(), (pvid->pending_ints&0x20)>>5, SekPc);
             goto update_irq;
           case 0x05:
           case 0x06:
@@ -982,32 +998,44 @@ PICO_INTERNAL_ASM void PicoVideoWrite(u32 a,unsigned short d)
           case 0x0c:
             // renderers should update their palettes if sh/hi mode is changed
             if ((d^dold)&8) Pico.m.dirtyPal = 1;
+            if ((d^dold)&1) {
+              PicoVideoFIFOMode(pvid->reg[1]&0x40, d & 1);
+              Pico.est.rendstatus |= PDRAW_DIRTY_SPRITES;
+            }
             break;
           default:
             return;
         }
-        SATaddr = ((pvid->reg[5]&0x7f) << 9) | ((pvid->reg[6]&0x20) << 11);
-        SATmask = ~0x1ff;
-        if (pvid->reg[12]&1)
-          SATaddr &= ~0x200, SATmask &= ~0x200; // H40, zero lowest SAT bit
-        //elprintf(EL_STATUS, "spritep moved to %04x", SATaddr);
+        if (Pico.est.rendstatus & PDRAW_DIRTY_SPRITES) {
+          SATaddr = ((pvid->reg[5]&0x7f) << 9) | ((pvid->reg[6]&0x20) << 11);
+          SATmask = ~0x1ff;
+          if (pvid->reg[12]&1)
+            SATaddr &= ~0x200, SATmask &= ~0x200; // H40, zero lowest SAT bit
+          //elprintf(EL_STATUS, "spritep moved to %04x", SATaddr);
+        }
         return;
 
 update_irq:
 #ifndef EMU_CORE_DEBUG
-        // update IRQ level
-        if (!SekShouldInterrupt()) // hack
+        // update IRQ level; TODO hack, still fire irq if disabling now
+        if (!SekShouldInterrupt() || SekIrqLevel < pvid->hint_irq)
         {
           int lines, pints, irq = 0;
           lines = (pvid->reg[1] & 0x20) | (pvid->reg[0] & 0x10);
           pints = pvid->pending_ints & lines;
                if (pints & 0x20) irq = 6;
           else if (pints & 0x10) irq = pvid->hint_irq;
-          if (SekIrqLevel < irq)
+
+          if (irq) {
+            // VDP irqs have highest prio, just overwrite old level
             SekInterrupt(irq); // update line
 
-          // this is broken because cost of current insn isn't known here
-          if (irq) SekEndRun(21); // make it delayed
+            // TODO this is broken because cost of current insn isn't known here
+            SekEndRun(21); // make it delayed
+          } else if (SekIrqLevel >= pvid->hint_irq) {
+            // no VDP irq, query lower irqs
+            SekInterrupt(PicoIn.AHW & PAHW_PICO ? PicoPicoIrqAck(0) : 0);
+          }
         }
 #endif
       }
@@ -1092,10 +1120,28 @@ PICO_INTERNAL_ASM u32 PicoVideoRead(u32 a)
     unsigned int c;
     u32 d;
 
+    // invalidate cached value if on a new scanline
+    if ((pv->status & PVS_HVLATCH) && pv->hv_latch >> 8 != Pico.m.scanline)
+      pv->status &= ~PVS_HVLATCH;
+
     c = SekCyclesDone() - Pico.t.m68c_line_start;
-    if (pv->reg[0]&2)
-         d = pv->hv_latch;
-    else d = VdpFIFO.fifo_hcounts[c/clkdiv] | (pv->v_counter << 8);
+    if (pv->reg[0]&2) // latched value from TH transition
+      d = pv->hv_latch;
+    else if (pv->status & PVS_HVLATCH) { // cached value from light gun
+      c += 28; // estimate for 68k insn time
+      // return cached value, but add in irq latency; since the irq was fired
+      // at the start of the line, the latency in cycles already is in c.
+      d = pv->hv_latch & 0xff;
+      if (pv->reg[12]&1) { // H40/H32?
+        c = DIVQ32(c * slots40, slcpu); // slots since line start
+        if (d < gapstart40 && d+c >= gapstart40) d += gapend40-gapstart40;
+      } else {
+        c = DIVQ32(c * slots32, slcpu);
+        if (d < gapstart32 && d+c >= gapstart32) d += gapend32-gapstart32;
+      }
+      d = (pv->hv_latch & 0xff00) | ((d+c) & 0x00ff);
+    } else // freerunning
+      d = VdpFIFO.fifo_hcounts[c/clkdiv] | (pv->v_counter << 8);
 
     elprintf(EL_HVCNT, "hv: %02x %02x [%u] @ %06x", d, pv->v_counter, SekCyclesDone(), SekPc);
     return d;
@@ -1145,8 +1191,11 @@ unsigned char PicoVideoRead8CtlL(int is_from_z80)
 
 unsigned char PicoVideoRead8HV_H(int is_from_z80)
 {
-  elprintf(EL_HVCNT, "vcounter: %02x [%u] @ %06x", Pico.video.v_counter, SekCyclesDone(), SekPc);
-  return Pico.video.v_counter;
+  u32 d = Pico.video.v_counter;
+  if (Pico.video.reg[0]&2)
+    d = Pico.video.hv_latch >> 8;
+  elprintf(EL_HVCNT, "vcounter: %02x [%u] @ %06x", d, SekCyclesDone(), SekPc);
+  return d;
 }
 
 // FIXME: broken
@@ -1160,6 +1209,49 @@ unsigned char PicoVideoRead8HV_L(int is_from_z80)
   return d;
 }
 
+void PicoVideoReset(void)
+{
+  struct PicoVideo *pv = &Pico.video;
+
+  pv->pending_ints=0;
+  pv->reg[1] &= ~0x40; // TODO verify display disabled after reset
+  pv->reg[10] = 0xff; // HINT is turned off after reset
+  pv->status = 0x3428 | Pico.m.pal; // 'always set' bits | vblank | collision | pal
+
+  memset(&VdpFIFO, 0, sizeof(VdpFIFO));
+  Pico.m.dirtyPal = 1;
+
+  PicoDrawBgcDMA(NULL, 0, 0, 0, 0);
+  PicoVideoFIFOMode(pv->reg[1]&0x40, pv->reg[12]&1);
+}
+
+// Process triggering of the !HL pin at pixel x,y of screen
+void PicoVideoTriggerTH(int x, int y)
+{
+  struct PicoVideo *pv = &Pico.video;
+  int slot;
+
+  // TODO x/y scaling for V28/V30, H32/H40, soft/hardscaling?
+  if (!(pv->reg[12]&1)) { // H32, convert from 320 to 256 px per line
+    slot = DIVQ32(x * 256, 320) / 2;
+    if (slot >= gapstart32) slot += gapend32-gapstart32;
+  } else {
+    slot = x / 2;
+    if (slot >= gapstart40) slot += gapend40-gapstart40;
+  }
+
+  // This is only called at the start of a scanline, hence the irq isn't
+  // cycle correct. There are games reading hv unlatched in the irq handler,
+  // so latch mouse here, and fix the value up when the irq handler reads it.
+  pv->hv_latch = (slot & 0xff) | (y << 8);
+  pv->status |= PVS_HVLATCH;
+
+  // irq here for speed reasons. It's early, but that shouldn't matter much.
+  // TODO is irq really lost if another irq is already pending?
+  if ((Pico.video.reg[11]&8) && SekIrqLevel < 2)
+    SekInterrupt(2);
+}
+
 void PicoVideoCacheSAT(int load)
 {
   struct PicoVideo *pv = &Pico.video;
@@ -1180,29 +1272,44 @@ void PicoVideoCacheSAT(int load)
   Pico.est.rendstatus |= PDRAW_DIRTY_SPRITES;
 }
 
-void PicoVideoSave(void)
-{
-  struct VdpFIFO *vf = &VdpFIFO;
-  struct PicoVideo *pv = &Pico.video;
-  int l, x;
+#include <stddef.h>
 
-  // account for all outstanding xfers XXX kludge, entry attr's not saved
-  pv->fifo_cnt = pv->fifo_bgcnt = 0;
-  for (l = vf->fifo_ql, x = vf->fifo_qx + l-1; l > 0; l--, x--) {
-    int cnt = (vf->fifo_queue[x&7] >> 3);
-    if (vf->fifo_queue[x&7] & FQ_BGDMA)
-      pv->fifo_bgcnt += cnt;
-    else
-      pv->fifo_cnt += cnt;
-  }
+int PicoVideoSave(void *buf)
+{
+  u8 *bp = buf;
+  int i;
+
+  // FIFO stuff
+  memcpy(bp, &VdpFIFO, offsetof(struct VdpFIFO, fifo_slot));
+  bp += offsetof(struct VdpFIFO, fifo_slot);
+
+  // SAT Cache
+  for (i = 0; i < 80; i++, bp += sizeof(u32))
+    memcpy(bp, VdpSATCache+2*i, sizeof(u32));
+
+  return bp - (u8 *)buf;
 }
 
-void PicoVideoLoad(void)
+void PicoVideoLoad(void *buf, int len)
 {
   struct VdpFIFO *vf = &VdpFIFO;
   struct PicoVideo *pv = &Pico.video;
   int b = pv->type == 1;
 
+  SATaddr = ((pv->reg[5]&0x7f) << 9) | ((pv->reg[6]&0x20) << 11);
+  SATmask = ~0x1ff;
+  if (pv->reg[12]&1)
+    SATaddr &= ~0x200, SATmask &= ~0x200; // H40, zero lowest SAT bit
+
+  if (len) {
+    int i;
+    if (len >= offsetof(struct VdpFIFO, fifo_slot))
+      memcpy(&VdpFIFO, buf, offsetof(struct VdpFIFO, fifo_slot));
+    for (i = 0; i < 80; i++)
+      memcpy(VdpSATCache+2*i, buf + offsetof(struct VdpFIFO, fifo_slot) + 4*i, sizeof(u32));
+    return;
+  }
+
   // convert former dma_xfers (why was this in PicoMisc anyway?)
   if (Pico.m.dma_xfers) {
     pv->fifo_cnt = Pico.m.dma_xfers << b;
@@ -1213,10 +1320,11 @@ void PicoVideoLoad(void)
   vf->fifo_ql = vf->fifo_qx = vf->fifo_total = 0;
   if (pv->fifo_cnt) {
     int wc = pv->fifo_cnt;
-    pv->status |= PVS_CPUWR;
     vf->fifo_total = (wc+b) >> b;
     vf->fifo_queue[vf->fifo_qx + vf->fifo_ql] = (wc << 3) | b | FQ_FGDMA;
     vf->fifo_ql ++;
+    if (vf->fifo_total > 4 && !(pv->status & (PVS_CPUWR|PVS_CPURD)))
+      pv->status |= PVS_CPUWR;
   }
   if (pv->fifo_bgcnt) {
     int wc = pv->fifo_bgcnt;
@@ -1225,8 +1333,7 @@ void PicoVideoLoad(void)
     vf->fifo_queue[vf->fifo_qx + vf->fifo_ql] = (wc << 3)     | FQ_BGDMA;
     vf->fifo_ql ++;
   }
-  if (vf->fifo_ql)
-    pv->status |= SR_DMA;
   PicoVideoCacheSAT(1);
+  vf->fifo_maxslot = 0;
 }
 // vim:shiftwidth=2:ts=2:expandtab
