@@ -3,7 +3,7 @@
  * (C) notaz, 2013
  * (C) aliaspider, 2016
  * (C) Daniel De Matteis, 2013
- * (C) kub, 2020
+ * (C) irixxxx, 2020-2024
  *
  * This work is licensed under the terms of MAME license.
  * See COPYING file in the top-level directory.
@@ -37,6 +37,7 @@
 
 #if defined(RENDER_GSKIT_PS2)
 #include <malloc.h>
+#include <kernel.h>
 #include "libretro-common/include/libretro_gskit_ps2.h"
 #include "ps2/asm.h"
 #else
@@ -123,6 +124,8 @@ static int vout_ghosting = 0;
 static bool libretro_update_av_info = false;
 static bool libretro_update_geometry = false;
 
+static unsigned short libretro_mem_AHW = 0;
+
 #if defined(RENDER_GSKIT_PS2)
 #define VOUT_8BIT_WIDTH 328
 #define VOUT_8BIT_HEIGHT 256
@@ -223,17 +226,18 @@ void cache_flush_d_inval_i(void *start, void *end)
 {
 #ifdef __arm__
    size_t len = (char *)end - (char *)start;
-   (void)len;
 #if defined(__BLACKBERRY_QNX__)
-   msync(start, end - start, MS_SYNC | MS_CACHE_ONLY | MS_INVALIDATE_ICACHE);
+   msync(start, len, MS_SYNC | MS_CACHE_ONLY | MS_INVALIDATE_ICACHE);
 #elif defined(__MACH__)
    sys_dcache_flush(start, len);
    sys_icache_invalidate(start, len);
 #elif defined(_3DS)
+   (void)len;
    ctr_flush_invalidate_cache();
 #elif defined(VITA)
    sceKernelSyncVMDomain(sceBlock, start, len);
 #else
+   (void)len;
    __clear_cache(start, end);
 #endif
 #endif
@@ -550,7 +554,7 @@ void emu_video_mode_change(int start_line, int line_count, int start_col, int co
    vm_current_start_col = start_col;
    vm_current_col_count = col_count;
 
-   // 8bit renderes create a 328x256 CLUT image, while 16bit creates 320x240 RGB
+   // 8bit renderers create a 328x256 CLUT image, 16bit creates 320x240 RGB
 #if defined(RENDER_GSKIT_PS2)
    // calculate the borders of the real image inside the picodrive image
    vout_width = (vout_16bit ? VOUT_MAX_WIDTH : VOUT_8BIT_WIDTH);
@@ -1009,9 +1013,11 @@ static bool disk_set_image_index(unsigned int index)
             disks[index].fname);
 
    ret = -1;
-   cd_type = PicoCdCheck(disks[index].fname, NULL);
-   if (cd_type >= 0 && cd_type != CT_UNKNOWN)
-      ret = cdd_load(disks[index].fname, cd_type);
+   if (PicoIn.AHW & PAHW_MCD) {
+     cd_type = PicoCdCheck(disks[index].fname, NULL);
+     if (cd_type >= 0 && cd_type != CT_UNKNOWN)
+        ret = cdd_load(disks[index].fname, cd_type);
+   }
    if (ret != 0) {
       if (log_cb)
          log_cb(RETRO_LOG_ERROR, "Load failed, invalid CD image?\n");
@@ -1284,22 +1290,6 @@ static const char *find_bios(int *region, const char *cd_fname)
    int i, count;
    FILE *f = NULL;
 
-   // look for MSU.MD rom file. XXX another extension list? ugh...
-   static const char *md_exts[] = { "gen", "smd", "md", "32x" };
-   char *ext = strrchr(cd_fname, '.');
-   int extpos = ext ? ext-cd_fname : strlen(cd_fname);
-   strcpy(path, cd_fname);
-   path[extpos++] = '.';
-   for (i = 0; i < ARRAY_SIZE(md_exts); i++) {
-      strcpy(path+extpos, md_exts[i]);
-      f = fopen(path, "rb");
-      if (f != NULL) {
-         log_cb(RETRO_LOG_INFO, "found MSU rom: %s\n", path);
-         fclose(f);
-         return path;
-      }
-   }
-
    if (*region == 4) { // US
       files = biosfiles_us;
       count = sizeof(biosfiles_us) / sizeof(char *);
@@ -1336,6 +1326,31 @@ static const char *find_bios(int *region, const char *cd_fname)
    return NULL;
 }
 
+static const char *find_msu(const char *cd_fname)
+{
+   static char path[256];
+   FILE *f = NULL;
+   int i;
+
+   // look for MSU.MD rom file. XXX another extension list? ugh...
+   static const char *md_exts[] = { "gen", "smd", "md", "32x" };
+   char *ext = strrchr(cd_fname, '.');
+   int extpos = ext ? ext-cd_fname : strlen(cd_fname);
+   strcpy(path, cd_fname);
+   path[extpos++] = '.';
+   for (i = 0; i < ARRAY_SIZE(md_exts); i++) {
+      strcpy(path+extpos, md_exts[i]);
+      f = fopen(path, "rb");
+      if (f != NULL) {
+         log_cb(RETRO_LOG_INFO, "found MSU rom: %s\n", path);
+         fclose(f);
+         return path;
+      }
+   }
+
+   return NULL;
+}
+
 static void set_memory_maps(void)
 {
    if (PicoIn.AHW & PAHW_MCD)
@@ -1344,16 +1359,34 @@ static void set_memory_maps(void)
       const uint64_t mem = RETRO_MEMDESC_SYSTEM_RAM;
       struct retro_memory_map mmaps;
       struct retro_memory_descriptor descs[] = {
-         { mem, PicoMem.ram,        0,           0xFF0000, 0, 0, 0x10000, "68KRAM" },
+         { mem, PicoMem.ram,           0,           0xFF0000, 0, 0, 0x10000, "68KRAM" },
          /* virtual address using SCD_BIT so all 512M of prg_ram can be accessed */
          /* at address $80020000 */
-         { mem, Pico_mcd->prg_ram,  0, SCD_BIT | 0x020000, 0, 0, 0x80000, "PRGRAM" },
+         { mem, Pico_mcd->prg_ram,     0, SCD_BIT | 0x020000, 0, 0, 0x80000, "PRGRAM" },
+         { mem, Pico_mcd->word_ram2M,  0,           0x200000, 0, 0, 0x40000, "WORDRAM" },
       };
 
       mmaps.descriptors = descs;
       mmaps.num_descriptors = sizeof(descs) / sizeof(descs[0]);
       environ_cb(RETRO_ENVIRONMENT_SET_MEMORY_MAPS, &mmaps);
    }
+   else if (PicoIn.AHW & PAHW_32X)
+   {
+      const uint64_t mem = RETRO_MEMDESC_SYSTEM_RAM;
+      const uint64_t sav = RETRO_MEMDESC_SAVE_RAM;
+      struct retro_memory_map mmaps;
+      struct retro_memory_descriptor descs[] = {
+         { mem, PicoMem.ram,        0, 0x00FF0000, 0, 0, 0x10000,      "68KRAM" },
+         { mem, Pico32xMem->sdram,  0, 0x06000000, 0, 0, 0x40000,      "SDRAM" },
+         { sav, Pico.sv.data,       0, 0x00000000, 0, 0, Pico.sv.size, "CARTRAM" },
+      };
+
+      mmaps.descriptors = descs;
+      mmaps.num_descriptors = sizeof(descs) / sizeof(descs[0]);
+      environ_cb(RETRO_ENVIRONMENT_SET_MEMORY_MAPS, &mmaps);
+   }
+
+   libretro_mem_AHW = PicoIn.AHW;
 }
 
 bool retro_load_game(const struct retro_game_info *info)
@@ -1584,7 +1617,7 @@ bool retro_load_game(const struct retro_game_info *info)
    make_system_path(carthw_path, sizeof(carthw_path), "carthw", ".cfg");
 
    media_type = PicoLoadMedia(content_path, content_data, content_size,
-         carthw_path, find_bios, NULL);
+         carthw_path, find_bios, find_msu, NULL);
 
    disk_current_index = cd_index;
 
@@ -1652,6 +1685,7 @@ bool retro_load_game_special(unsigned game_type, const struct retro_game_info *i
 
 void retro_unload_game(void)
 {
+    libretro_mem_AHW = 0;
 }
 
 unsigned retro_get_region(void)
@@ -1661,16 +1695,16 @@ unsigned retro_get_region(void)
 
 void *retro_get_memory_data(unsigned type)
 {
-   uint8_t* data;
+   void *data;
 
    switch(type)
    {
       case RETRO_MEMORY_SAVE_RAM:
          /* Note: MCD RAM cart uses Pico.sv.data */
-         if ((PicoIn.AHW & PAHW_MCD) &&
+         if ((PicoIn.AHW & PAHW_MCD) && Pico.romsize == 0 &&
                !(PicoIn.opt & POPT_EN_MCD_RAMCART))
             data = Pico_mcd->bram;
-         else
+         else // TODO: handle copying to/from bram somewhere
             data = Pico.sv.data;
          break;
       case RETRO_MEMORY_SYSTEM_RAM:
@@ -1701,7 +1735,7 @@ size_t retro_get_memory_size(unsigned type)
    switch(type)
    {
       case RETRO_MEMORY_SAVE_RAM:
-         if (PicoIn.AHW & PAHW_MCD)
+         if ((PicoIn.AHW & PAHW_MCD) && Pico.romsize == 0)
          {
             if (PicoIn.opt & POPT_EN_MCD_RAMCART)
                return 0x12000;
@@ -2249,9 +2283,15 @@ void run_events_pico(unsigned int events)
     }
     if (events & (1 << RETRO_DEVICE_ID_JOYPAD_R)) {
 	PicoPicohw.page++;
-	if (PicoPicohw.page > 6)
-	    PicoPicohw.page = 6;
-	emu_status_msg("Page %i", PicoPicohw.page);
+	if (PicoPicohw.page > 7)
+	    PicoPicohw.page = 7;
+        if (PicoPicohw.page == 7) {
+            // Used in games that require the Keyboard Pico peripheral
+            emu_status_msg("Test Page");
+        }
+        else {
+            emu_status_msg("Page %i", PicoPicohw.page);
+        }
     }
     if (events & (1 << RETRO_DEVICE_ID_JOYPAD_X)) {
         if (pico_inp_mode == 2) {
@@ -2262,7 +2302,7 @@ void run_events_pico(unsigned int events)
             emu_status_msg("Input: Pen on Pad");
         }
     }
-    if (events & (1 << RETRO_DEVICE_ID_JOYPAD_SELECT)) {
+    if (events & (1 << RETRO_DEVICE_ID_JOYPAD_Y)) {
         if (pico_inp_mode == 1) {
             pico_inp_mode = 0;
             emu_status_msg("Input: D-Pad");
@@ -2312,6 +2352,9 @@ void retro_run(void)
    int pad, i, padcount;
    static void *buff;
 
+   if (PicoIn.AHW != libretro_mem_AHW)
+      set_memory_maps();
+
    PicoIn.skipFrame = 0;
 
    if (environ_cb(RETRO_ENVIRONMENT_GET_VARIABLE_UPDATE, &updated) && updated)
@@ -2355,7 +2398,7 @@ void retro_run(void)
    if (PicoIn.AHW == PAHW_PICO) {
        uint16_t ev = input[0] &
              ((1 << RETRO_DEVICE_ID_JOYPAD_L) | (1 << RETRO_DEVICE_ID_JOYPAD_R) |
-              (1 << RETRO_DEVICE_ID_JOYPAD_X) | (1 << RETRO_DEVICE_ID_JOYPAD_SELECT) |
+              (1 << RETRO_DEVICE_ID_JOYPAD_X) | (1 << RETRO_DEVICE_ID_JOYPAD_Y) |
               (1 << RETRO_DEVICE_ID_JOYPAD_START));
        uint16_t new_ev = ev & ~pico_events;
        pico_events = ev;
@@ -2557,7 +2600,7 @@ void retro_init(void)
       | POPT_EN_MCD_PCM|POPT_EN_MCD_CDDA|POPT_EN_MCD_GFX
       | POPT_EN_32X|POPT_EN_PWM
       | POPT_ACC_SPRITES|POPT_DIS_32C_BORDER;
-#ifdef __arm__
+#ifdef DRC_SH2
 #ifdef _3DS
    if (ctr_svchack_successful)
 #endif
